from logging import getLogger
from multiprocessing import cpu_count
from pathlib import Path
from time import perf_counter
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from darwin.datatypes import AnnotationFile
from darwin.item import DatasetItem

Unknown = Any  # type: ignore

from tqdm import tqdm

if TYPE_CHECKING:
    from darwin.client import Client
    from darwin.dataset import RemoteDataset

import deprecation
from rich.console import Console
from rich.progress import track
from rich.theme import Theme

import darwin.datatypes as dt
from darwin.datatypes import PathLike
from darwin.exceptions import IncompatibleOptions, RequestEntitySizeExceeded
from darwin.utils import flatten_list, secure_continue_request
from darwin.version import __version__

try:
    from mpire import WorkerPool

    MPIRE_AVAILABLE = True
except ImportError:
    MPIRE_AVAILABLE = False

# Classes missing import support on backend side
UNSUPPORTED_CLASSES = ["string", "graph"]

DEPRECATION_MESSAGE = """

This function is going to be turned into private. This means that breaking
changes in its interface and implementation are to be expected. We encourage using ``import_annotations``
instead of calling this low-level function directly.

"""


@deprecation.deprecated(  # type:ignore
    deprecated_in="0.7.12",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_main_annotations_lookup_table(annotation_classes: List[Dict[str, Unknown]]) -> Dict[str, Unknown]:
    MAIN_ANNOTATION_TYPES = [
        "bounding_box",
        "cuboid",
        "ellipse",
        "keypoint",
        "line",
        "link",
        "polygon",
        "skeleton",
        "tag",
        "string",
        "table",
        "graph",
    ]
    lookup: Dict[str, Unknown] = {}
    for cls in annotation_classes:
        for annotation_type in cls["annotation_types"]:
            if annotation_type in MAIN_ANNOTATION_TYPES:
                if annotation_type not in lookup:
                    lookup[annotation_type] = {}
                lookup[annotation_type][cls["name"]] = cls["id"]

    return lookup


@deprecation.deprecated(  # type:ignore
    deprecated_in="0.7.12",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def find_and_parse(
    importer: Callable[[Path], Union[List[dt.AnnotationFile], dt.AnnotationFile, None]],
    file_paths: List[PathLike],
    console: Optional[Console] = None,
    use_multi_cpu: bool = True,
    cpu_limit: int = 1,
) -> Optional[Iterable[dt.AnnotationFile]]:
    is_console = console is not None

    logger = getLogger(__name__)

    def perf_time(reset: bool = False) -> Generator[float, float, None]:
        start = perf_counter()
        yield start
        while True:
            if reset:
                start = perf_counter()
            yield perf_counter() - start

    def maybe_console(*args: Union[str, int, float]) -> None:
        if console is not None:
            console.print(*[f"[{str(next(perf_time()))} seconds elapsed]", *args])
        else:
            logger.info(*[f"[{str(next(perf_time()))}]", *args])

    maybe_console("Parsing files... ")

    files: List[Path] = _get_files_for_parsing(file_paths)

    maybe_console(f"Found {len(files)} files")

    if use_multi_cpu and MPIRE_AVAILABLE and cpu_limit > 1:
        maybe_console(f"Using multiprocessing with {cpu_limit} workers")
        try:
            with WorkerPool(cpu_limit) as pool:
                parsed_files = pool.map(importer, files, progress_bar=is_console)
        except KeyboardInterrupt:
            maybe_console("Keyboard interrupt. Stopping.")
            return None
        except Exception as e:
            maybe_console(f"Error: {e}")
            return None

    else:
        maybe_console("Using single CPU")
        parsed_files = list(map(importer, tqdm(files) if is_console else files))

    maybe_console("Finished.")

    if not isinstance(parsed_files, list):
        parsed_files = [parsed_files]

    parsed_files = [f for f in parsed_files if f is not None]
    return parsed_files


def _get_files_for_parsing(file_paths: List[PathLike]) -> List[Path]:
    packed_files = [filepath.glob("**/*") if filepath.is_dir() else [filepath] for filepath in map(Path, file_paths)]
    return [file for files in packed_files for file in files]


@deprecation.deprecated(  # type:ignore
    deprecated_in="0.7.12",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_attribute_lookup(dataset: "RemoteDataset") -> Dict[str, Unknown]:
    attributes: List[Dict[str, Unknown]] = dataset.fetch_remote_attributes()
    lookup: Dict[str, Unknown] = {}
    for attribute in attributes:
        class_id = attribute["class_id"]
        if class_id not in lookup:
            lookup[class_id] = {}
        lookup[class_id][attribute["name"]] = attribute["id"]
    return lookup


@deprecation.deprecated(  # type:ignore
    deprecated_in="0.7.12",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def get_remote_files(
    dataset: "RemoteDataset", filenames: List[str], chunk_size: int = 100
) -> Dict[str, Tuple[int, str]]:
    """
    Fetches remote files from the datasets in chunks; by default 100 filenames at a time.

    The output is a two-element tuple of:
    - file ID
    - the name of the first slot for V2 items, or '0' for V1 items

    Fetching slot name is necessary here to avoid double-trip to Api downstream for remote files.
    """
    remote_files = {}
    for i in range(0, len(filenames), chunk_size):
        chunk = filenames[i : i + chunk_size]
        for remote_file in dataset.fetch_remote_files(
            {"types": "image,playback_video,video_frame", "filenames": chunk}
        ):
            slot_name = _get_slot_name(remote_file)
            remote_files[remote_file.full_path] = (remote_file.id, slot_name)
    return remote_files


def _get_slot_name(remote_file: DatasetItem) -> str:
    slot = next(iter(remote_file.slots), {"slot_name": "0"})
    if slot:
        return slot["slot_name"]
    else:
        return "0"


def _resolve_annotation_classes(
    local_annotation_classes: List[dt.AnnotationClass],
    classes_in_dataset: Dict[str, Unknown],
    classes_in_team: Dict[str, Unknown],
) -> Tuple[Set[dt.AnnotationClass], Set[dt.AnnotationClass]]:
    local_classes_not_in_dataset: Set[dt.AnnotationClass] = set()
    local_classes_not_in_team: Set[dt.AnnotationClass] = set()

    for local_cls in local_annotation_classes:
        local_annotation_type = local_cls.annotation_internal_type or local_cls.annotation_type
        # Only add the new class if it doesn't exist remotely already
        if local_annotation_type in classes_in_dataset and local_cls.name in classes_in_dataset[local_annotation_type]:
            continue

        # Only add the new class if it's not included in the list of the missing classes already
        if local_cls.name in [missing_class.name for missing_class in local_classes_not_in_dataset]:
            continue
        if local_cls.name in [missing_class.name for missing_class in local_classes_not_in_team]:
            continue

        if local_annotation_type in classes_in_team and local_cls.name in classes_in_team[local_annotation_type]:
            local_classes_not_in_dataset.add(local_cls)
        else:
            local_classes_not_in_team.add(local_cls)

    return local_classes_not_in_dataset, local_classes_not_in_team


def import_annotations(
    dataset: "RemoteDataset",
    importer: Callable[[Path], Union[List[dt.AnnotationFile], dt.AnnotationFile, None]],
    file_paths: List[PathLike],
    append: bool,
    class_prompt: bool = True,
    delete_for_empty: bool = False,
    import_annotators: bool = False,
    import_reviewers: bool = False,
    use_multi_cpu: bool = False,  # Set to False to give time to resolve MP behaviours
    cpu_limit: Optional[int] = None,  # 0 because it's set later in logic
) -> None:
    """
    Imports the given given Annotations into the given Dataset.

    Parameters
    ----------
    dataset : RemoteDataset
        Dataset where the Annotations will be imported to.
    importer : Callable[[Path], Union[List[dt.AnnotationFile], dt.AnnotationFile, None]]
        Parsing module containing the logic to parse the given Annotation files given in
        ``files_path``. See ``importer/format`` for a list of out of supported parsers.
    file_paths : List[PathLike]
        A list of ``Path``'s or strings containing the Annotations we wish to import.
    append : bool
        If ``True`` appends the given annotations to the datasets. If ``False`` will override them.
        Incompatible with ``delete-for-empty``.
    class_prompt : bool
        If ``False`` classes will be created and added to the datasets without requiring a user's prompt.
    delete_for_empty : bool, default: False
        If ``True`` will use empty annotation files to delete all annotations from the remote file.
        If ``False``, empty annotation files will simply be skipped.
        Only works for V2 datasets.
        Incompatible with ``append``.
    import_annotators : bool, default: False
        If ``True`` it will import the annotators from the files to the dataset, if available.
        If ``False`` it will not import the annotators.
    import_reviewers : bool, default: False
        If ``True`` it will import the reviewers from the files to the dataset, if .
        If ``False`` it will not import the reviewers.
    use_multi_cpu : bool, default: True
        If ``True`` will use multiple available CPU cores to parse the annotation files.
        If ``False`` will use only the current Python process, which runs in one core.
        Processing using multiple cores is faster, but may slow down a machine also running other processes.
        Processing with one core is slower, but will run well alongside other processes.
    cpu_limit : int, default: 2 less than total cpu count
        The maximum number of CPU cores to use when ``use_multi_cpu`` is ``True``.
        If ``cpu_limit`` is greater than the number of available CPU cores, it will be set to the number of available cores.
        If ``cpu_limit`` is less than 1, it will be set to CPU count - 2.
        If ``cpu_limit`` is omitted, it will be set to CPU count - 2.

    Raises
    -------
    ValueError

        - If ``file_paths`` is not a list.
        - If the application is unable to fetch any remote classes.
        - If the application was unable to find/parse any annotation files.
        - If the application was unable to fetch remote file list.

    IncompatibleOptions

        - If both ``append`` and ``delete_for_empty`` are specified as ``True``.
    """
    console = Console(theme=_console_theme())

    if append and delete_for_empty:
        raise IncompatibleOptions(
            "The options 'append' and 'delete_for_empty' cannot be used together. Use only one of them."
        )

    cpu_limit, use_multi_cpu = _get_multi_cpu_settings(cpu_limit, cpu_count(), use_multi_cpu)
    if use_multi_cpu:
        console.print(f"Using {cpu_limit} CPUs for parsing...", style="info")
    else:
        console.print("Using 1 CPU for parsing...", style="info")

    if not isinstance(file_paths, list):
        raise ValueError(f"file_paths must be a list of 'Path' or 'str'. Current value: {file_paths}")

    console.print("Fetching remote class list...", style="info")
    team_classes: List[dt.DictFreeForm] = dataset.fetch_remote_classes(True)
    if not team_classes:
        raise ValueError("Unable to fetch remote class list.")

    if delete_for_empty and dataset.version == 1:
        console.print(
            f"The '--delete-for-empty' flag only works for V2 datasets. '{dataset.name}' is a V1 dataset. Ignoring flag.",
            style="warning",
        )

    classes_in_dataset: dt.DictFreeForm = build_main_annotations_lookup_table(
        [cls for cls in team_classes if cls["available"]]
    )
    classes_in_team: dt.DictFreeForm = build_main_annotations_lookup_table(
        [cls for cls in team_classes if not cls["available"]]
    )
    attributes = build_attribute_lookup(dataset)

    console.print("Retrieving local annotations ...", style="info")
    local_files = []
    local_files_missing_remotely = []

    # ! Other place we can use multiprocessing - hard to pass in the importer though
    maybe_parsed_files: Optional[Iterable[dt.AnnotationFile]] = find_and_parse(
        importer, file_paths, console, use_multi_cpu, cpu_limit
    )

    if not maybe_parsed_files:
        raise ValueError("Not able to parse any files.")

    parsed_files: List[AnnotationFile] = flatten_list(list(maybe_parsed_files))

    filenames: List[str] = [parsed_file.filename for parsed_file in parsed_files if parsed_file is not None]

    console.print("Fetching remote file list...", style="info")
    # This call will only filter by filename; so can return a superset of matched files across different paths
    # There is logic in this function to then include paths to narrow down to the single correct matching file
    remote_files: Dict[str, Tuple[int, str]] = dict()

    # Try to fetch files in large chunks; in case the filenames are too large and exceed the url size
    # retry in smaller chunks
    chunk_size = 100
    while chunk_size > 0:
        try:
            remote_files = get_remote_files(dataset, filenames, chunk_size)
            break
        except RequestEntitySizeExceeded:
            chunk_size -= 8
            if chunk_size <= 0:
                raise ValueError("Unable to fetch remote file list.")

    for parsed_file in parsed_files:
        if parsed_file.full_path not in remote_files:
            local_files_missing_remotely.append(parsed_file)
        else:
            local_files.append(parsed_file)

    console.print(f"{len(local_files) + len(local_files_missing_remotely)} annotation file(s) found.", style="info")
    if local_files_missing_remotely:
        console.print(f"{len(local_files_missing_remotely)} file(s) are missing from the dataset", style="warning")
        for local_file in local_files_missing_remotely:
            console.print(f"\t{local_file.path}: '{local_file.full_path}'", style="warning")

        if class_prompt and not secure_continue_request():
            return

    local_classes_not_in_dataset, local_classes_not_in_team = _resolve_annotation_classes(
        [annotation_class for file in local_files for annotation_class in file.annotation_classes],
        classes_in_dataset,
        classes_in_team,
    )

    console.print(f"{len(local_classes_not_in_team)} classes needs to be created.", style="info")
    console.print(
        f"{len(local_classes_not_in_dataset)} classes needs to be added to {dataset.identifier}", style="info"
    )

    missing_skeletons: List[dt.AnnotationClass] = list(filter(_is_skeleton_class, local_classes_not_in_team))
    missing_skeleton_names: str = ", ".join(map(_get_skeleton_name, missing_skeletons))
    if missing_skeletons:
        console.print(
            f"Found missing skeleton classes: {missing_skeleton_names}. Missing Skeleton classes cannot be created. Exiting now.",
            style="error",
        )
        return

    if local_classes_not_in_team:
        console.print("About to create the following classes", style="info")
        for missing_class in local_classes_not_in_team:
            console.print(
                f"\t{missing_class.name}, type: {missing_class.annotation_internal_type or missing_class.annotation_type}",
                style="info",
            )
        if class_prompt and not secure_continue_request():
            return
        for missing_class in local_classes_not_in_team:
            dataset.create_annotation_class(
                missing_class.name, missing_class.annotation_internal_type or missing_class.annotation_type
            )
    if local_classes_not_in_dataset:
        console.print(f"About to add the following classes to {dataset.identifier}", style="info")
        for cls in local_classes_not_in_dataset:
            dataset.add_annotation_class(cls)

    # Refetch classes to update mappings
    if local_classes_not_in_team or local_classes_not_in_dataset:
        maybe_remote_classes: List[dt.DictFreeForm] = dataset.fetch_remote_classes()
        if not maybe_remote_classes:
            raise ValueError("Unable to fetch remote classes.")

        remote_classes = build_main_annotations_lookup_table(maybe_remote_classes)
    else:
        remote_classes = build_main_annotations_lookup_table(team_classes)

    if dataset.version == 1:
        console.print("Importing annotations...\nEmpty annotations will be skipped.", style="info")
    elif dataset.version == 2 and delete_for_empty:
        console.print(
            "Importing annotations...\nEmpty annotation file(s) will clear all existing annotations in matching remote files.",
            style="info",
        )
    else:
        console.print(
            "Importing annotations...\nEmpty annotations will be skipped, if you want to delete annotations rerun with '--delete-for-empty'.",
            style="info",
        )

    # Need to re parse the files since we didn't save the annotations in memory
    for local_path in set(local_file.path for local_file in local_files):

        imported_files: Union[List[dt.AnnotationFile], dt.AnnotationFile, None] = importer(local_path)
        if imported_files is None:
            parsed_files = []
        elif not isinstance(imported_files, List):
            parsed_files = [imported_files]
        else:
            parsed_files = imported_files

        # remove files missing on the server
        missing_files = [missing_file.full_path for missing_file in local_files_missing_remotely]
        parsed_files = [parsed_file for parsed_file in parsed_files if parsed_file.full_path not in missing_files]

        files_to_not_track = [
            file_to_track
            for file_to_track in parsed_files
            if not file_to_track.annotations and (not delete_for_empty or dataset.version == 1)
        ]

        for file in files_to_not_track:
            console.print(f"{file.filename} has no annotations. Skipping upload...", style="warning")

        files_to_track = [file for file in parsed_files if file not in files_to_not_track]
        if files_to_track:
            _warn_unsupported_annotations(files_to_track)
            for parsed_file in track(files_to_track):

                image_id, default_slot_name = remote_files[parsed_file.full_path]

                errors, succes = _import_annotations(
                    dataset.client,
                    image_id,
                    remote_classes,
                    attributes,
                    parsed_file.annotations,  # type: ignore
                    default_slot_name,
                    dataset,
                    append,
                    delete_for_empty,
                    import_annotators,
                    import_reviewers,
                )

                if errors:
                    console.print(f"Errors importing {parsed_file.filename}", style="error")
                    for error in errors:
                        console.print(f"\t{error}", style="error")


def _get_multi_cpu_settings(cpu_limit: Optional[int], cpu_count: int, use_multi_cpu: bool) -> Tuple[int, bool]:
    if cpu_limit == 1 or cpu_count == 1 or not use_multi_cpu:
        return 1, False

    if cpu_limit is None:
        return max([cpu_count - 2, 2]), True

    return cpu_limit if cpu_limit <= cpu_count else cpu_count, True


def _warn_unsupported_annotations(parsed_files: List[AnnotationFile]) -> None:
    console = Console(theme=_console_theme())
    for parsed_file in parsed_files:
        skipped_annotations = []
        for annotation in parsed_file.annotations:
            if annotation.annotation_class.annotation_type in UNSUPPORTED_CLASSES:
                skipped_annotations.append(annotation)
        if len(skipped_annotations) > 0:
            types = set(map(lambda c: c.annotation_class.annotation_type, skipped_annotations))
            console.print(
                f"Import of annotation class types '{', '.join(types)}' is not yet supported. Skipping {len(skipped_annotations)} "
                + "annotations from '{parsed_file.full_path}'.\n",
                style="warning",
            )


def _is_skeleton_class(the_class: dt.AnnotationClass) -> bool:
    return (the_class.annotation_internal_type or the_class.annotation_type) == "skeleton"


def _get_skeleton_name(skeleton: dt.AnnotationClass) -> str:
    return skeleton.name


def _handle_subs(
    annotation: dt.Annotation, data: dt.DictFreeForm, annotation_class_id: str, attributes: Dict[str, dt.UnknownType]
) -> dt.DictFreeForm:
    for sub in annotation.subs:
        if sub.annotation_type == "text":
            data["text"] = {"text": sub.data}
        elif sub.annotation_type == "attributes":
            attributes_with_key = []
            for attr in sub.data:
                if annotation_class_id in attributes and attr in attributes[annotation_class_id]:
                    attributes_with_key.append(attributes[annotation_class_id][attr])
                else:
                    print(f"The attribute '{attr}' for class '{annotation.annotation_class.name}' was not imported.")

            data["attributes"] = {"attributes": attributes_with_key}
        elif sub.annotation_type == "instance_id":
            data["instance_id"] = {"value": sub.data}
        else:
            data[sub.annotation_type] = sub.data

    return data


def _handle_complex_polygon(annotation: dt.Annotation, data: dt.DictFreeForm) -> dt.DictFreeForm:
    if "complex_polygon" in data:
        del data["complex_polygon"]
        data["polygon"] = {"path": annotation.data["paths"][0], "additional_paths": annotation.data["paths"][1:]}
    return data


def _annotators_or_reviewers_to_payload(
    actors: List[dt.AnnotationAuthor], role: dt.AnnotationAuthorRole
) -> List[dt.DictFreeForm]:

    return [{"email": actor.email, "role": role.value} for actor in actors]


def _handle_reviewers(annotation: dt.Annotation, import_reviewers: bool) -> List[dt.DictFreeForm]:
    if import_reviewers:
        if annotation.reviewers:
            return _annotators_or_reviewers_to_payload(annotation.reviewers, dt.AnnotationAuthorRole.REVIEWER)
    return []


def _handle_annotators(annotation: dt.Annotation, import_annotators: bool) -> List[dt.DictFreeForm]:
    if import_annotators:
        if annotation.annotators:
            return _annotators_or_reviewers_to_payload(annotation.annotators, dt.AnnotationAuthorRole.ANNOTATOR)
    return []


def _get_annotation_data(
    annotation: dt.AnnotationLike, annotation_class_id: str, attributes: dt.DictFreeForm
) -> dt.DictFreeForm:
    annotation_class = annotation.annotation_class
    if isinstance(annotation, dt.VideoAnnotation):
        data = annotation.get_data(
            only_keyframes=True,
            post_processing=lambda annotation, data: _handle_subs(
                annotation, _handle_complex_polygon(annotation, data), annotation_class_id, attributes
            ),
        )
    else:
        data = {annotation_class.annotation_type: annotation.data}
        data = _handle_complex_polygon(annotation, data)
        data = _handle_subs(annotation, data, annotation_class_id, attributes)

    return data


def _handle_slot_names(annotation: dt.Annotation, dataset_version: int, default_slot_name: str) -> dt.Annotation:
    if not annotation.slot_names and dataset_version > 1:
        annotation.slot_names.extend([default_slot_name])

    return annotation


def _get_overwrite_value(append: bool) -> str:
    return "false" if append else "true"


def _import_annotations(
    client: "Client",  # TODO: This is unused, should it be?
    id: Union[str, int],
    remote_classes: dt.DictFreeForm,
    attributes: dt.DictFreeForm,
    annotations: List[dt.Annotation],
    default_slot_name: str,
    dataset: "RemoteDataset",
    append: bool,
    delete_for_empty: bool,  # TODO: This is unused, should it be?
    import_annotators: bool,
    import_reviewers: bool,
) -> Tuple[dt.ErrorList, dt.Success]:
    errors: dt.ErrorList = []
    success: dt.Success = dt.Success.SUCCESS

    serialized_annotations = []
    for annotation in annotations:
        annotation_class = annotation.annotation_class
        annotation_type = annotation_class.annotation_internal_type or annotation_class.annotation_type
        annotation_class_id: str = remote_classes[annotation_type][annotation_class.name]

        data = _get_annotation_data(annotation, annotation_class_id, attributes)

        actors: List[dt.DictFreeForm] = []
        actors.extend(_handle_annotators(annotation, import_annotators))
        actors.extend(_handle_reviewers(annotation, import_reviewers))

        # Insert the default slot name if not available in the import source
        annotation = _handle_slot_names(annotation, dataset.version, default_slot_name)

        serial_obj = {
            "annotation_class_id": annotation_class_id,
            "data": data,
            "context_keys": {"slot_names": annotation.slot_names},
        }

        if actors:
            serial_obj["actors"] = actors  # type: ignore

        serialized_annotations.append(serial_obj)

    payload: dt.DictFreeForm = {"annotations": serialized_annotations}
    payload["overwrite"] = _get_overwrite_value(append)

    try:
        dataset.import_annotation(id, payload=payload)
    except Exception as e:
        errors.append(e)
        success = dt.Success.FAILURE

    return errors, success


def _console_theme() -> Theme:
    return Theme(
        {"success": "bold green", "warning": "bold yellow", "error": "bold red", "info": "bold deep_sky_blue1"}
    )



from typing import Union

from darwin.doc_enum import DocEnum


class SortDirection(DocEnum):
    """
    The sorting direction of items.
    """

    ASCENDING = "asc", "Ascending sort order."
    DESCENDING = "desc", "Descending sort order."

    @classmethod
    def parse(cls, direction: str) -> "SortDirection":
        """
        Parses the given direction and returns the corresponding sort Enum.

        Parameters
        ----------
        direction: str
            The direction of the sorting order. Can be 'asc' or 'ascending', 'desc' or 'descending'.

        Returns
        -------
        SortDirection
            The Enum representing a sorting direction.

        Raises
        ------
        ValueError
            If the ``direction`` given is invalid.
        """
        normalized_direction = direction.lower()

        if cls._is_ascending(normalized_direction):
            return cls.ASCENDING
        if cls._is_descending(normalized_direction):
            return cls.DESCENDING

        raise ValueError(f"Invalid direction '{direction}', use 'asc' or 'ascending', 'desc' or 'descending'.")

    @staticmethod
    def _is_ascending(direction: str) -> bool:
        return direction == "asc" or direction == "ascending"

    @staticmethod
    def _is_descending(direction: str) -> bool:
        return direction == "desc" or direction == "descending"


class ItemSorter:
    """
    Represents sorting for list of items.

    Parameters
    ----------
    field : str
        The name of the field to be sorted.
    direction : SortDirection
        The direction of the sort.
    """

    def __init__(self, field: str, direction: SortDirection):
        self.field = field
        self.direction = direction

    @classmethod
    def parse(cls, sort: Union[str, "ItemSorter"]) -> "ItemSorter":
        """
        Parses the sorting given into an ItemSorter, capable of being used by Darwin.

        Parameters
        ----------
        sort : Union[str, ItemSorter]
            The sort order. If it is a ``str``, it will be parsed, otherwise it returns the
            ``ItemSorter``.

        Returns
        -------
        ItemSorter
            A parsed ``ItemSorter`` representing a sorting direction.

        Raises
        ------
        ValueError
            If the given sort parameter is invalid.
        """
        if isinstance(sort, ItemSorter):
            return sort

        if not cls._has_valid_format(sort):
            raise ValueError(
                f"Invalid sort parameter '{sort}'. Correct format is 'field:direction' where 'direction' is optional and defaults to 'asc', i.e. 'updated_at:asc' or just 'updated_at'."
            )

        if not cls._has_direction(sort):
            field = sort
            direction = "asc"
        else:
            field, direction = sort.split(":")

        if not cls._has_valid_field(field):
            raise ValueError(
                f"Invalid sort parameter '{field}', available sort fields: 'inserted_at', 'updated_at', 'file_size', 'filename', 'priority'."
            )

        return cls(field=field, direction=SortDirection.parse(direction))

    @staticmethod
    def _has_direction(sort: str) -> bool:
        return ":" in sort

    @staticmethod
    def _has_valid_format(sort_by: str) -> bool:
        return len(sort_by.split(":")) in [1, 2]

    @staticmethod
    def _has_valid_field(sort: str) -> bool:
        return sort in ["inserted_at", "updated_at", "file_size", "filename", "priority"]

    def __str__(self):
        return f"{self.field}:{self.direction.value}"



import concurrent.futures
import os
import time
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
)

import requests

from darwin.datatypes import PathLike
from darwin.doc_enum import DocEnum
from darwin.path_utils import construct_full_path
from darwin.utils import chunk

if TYPE_CHECKING:
    from darwin.client import Client
    from darwin.dataset import RemoteDataset
    from darwin.dataset.identifier import DatasetIdentifier

from abc import ABC, abstractmethod
from typing import Dict


class ItemPayload:
    """
    Represents an item's payload.

    Parameters
    ----------
    dataset_item_id : int
        The id of the dataset this item belongs to.
    filename : str
        The filename of where this ``ItemPayload``'s data is.
    path : str
        The path to ``filename``.
    reason : Optional[str], default: None
        A reason to upload this ``ItemPayload``.

    Attributes
    ----------
    dataset_item_id : int
        The id of the dataset this item belongs to.
    filename : str
        The filename of where this ``ItemPayload``'s data is.
    path : str
        The path to ``filename``.
    reason : Optional[str], default: None
        A reason to upload this ``ItemPayload``.
    """

    def __init__(
        self,
        *,
        dataset_item_id: int,
        filename: str,
        path: str,
        reason: Optional[str] = None,
        slots: Optional[any] = None,
    ):
        self.dataset_item_id = dataset_item_id
        self.filename = filename
        self.path = path
        self.reason = reason
        self.slots = slots

    @staticmethod
    def parse_v2(payload):
        if len(payload["slots"]) > 1:
            raise NotImplemented("multiple files support not yet implemented")
        slot = payload["slots"][0]
        return ItemPayload(
            dataset_item_id=payload.get("id", None),
            filename=payload["name"],
            path=payload["path"],
            reason=slot.get("reason", None),
            slots=payload["slots"],
        )

    @property
    def full_path(self) -> str:
        """The full ``Path`` (with filename inclduded) to the file."""
        return construct_full_path(self.path, self.filename)


class UploadStage(DocEnum):
    """
    The different stages of uploading a file.
    """

    REQUEST_SIGNATURE = 0, "First stage, when authentication is being performed."
    UPLOAD_TO_S3 = 1, "Second stage, when the file is being uploaded to S3."
    CONFIRM_UPLOAD_COMPLETE = 2, "Final stage, when we confirm the file was correctly uploaded."
    OTHER = 3, "If the stage of the upload process is unknown."


@dataclass
class UploadRequestError(Exception):
    """
    Error throw when uploading a file fails with an unrecoverable error.
    """

    #: The ``Path`` of the file being uploaded.
    file_path: Path

    #: The ``UploadStage`` when the  failure happened.
    stage: UploadStage

    #: The ``Exception`` that triggered this unrecoverable error.
    error: Optional[Exception] = None


class LocalFile:
    """
    Represents a file locally stored.

    Parameters
    ----------
    local_path : PathLike
        The ``Path`` of the file.
    kwargs : Any
        Data relative to this file. Can be anything.

    Attributes
    ----------
    local_path : PathLike
        The ``Path`` of the file.
    data : Dict[str, str]
        Dictionary with metadata relative to this file. It has the following format:

        .. code-block:: python

            {
                "filename": "a_filename",
                "path": "a path"
            }

        - ``data["filename"]`` will hold the value passed as ``filename`` from ``kwargs`` or default to ``self.local_path.name``
        - ``data["path"]`` will hold the value passed as ``path`` from ``kwargs`` or default to ``"/"``

    """

    def __init__(self, local_path: PathLike, **kwargs):
        self.local_path = Path(local_path)
        self.data = kwargs
        self._type_check(kwargs)

    def _type_check(self, args) -> None:
        self.data["filename"] = args.get("filename") or self.local_path.name
        self.data["path"] = args.get("path") or "/"

    def serialize(self):
        return {"files": [{"file_name": self.data["filename"], "slot_name": "0"}], "name": self.data["filename"]}

    def serialize_v2(self):
        optional_properties = ["tags", "fps", "as_frames", "extract_views"]
        slot = {"file_name": self.data["filename"], "slot_name": "0"}
        for optional_property in optional_properties:
            if optional_property in self.data:
                slot[optional_property] = self.data.get(optional_property)

        return {
            "slots": [slot],
            "name": self.data["filename"],
            "path": self.data["path"],
        }

    @property
    def full_path(self) -> str:
        """The full ``Path`` (with filename inclduded) to the file."""
        return construct_full_path(self.data["path"], self.data["filename"])


class FileMonitor(object):
    """
    Monitors the progress of a :class:``BufferedReader``.

    To use this monitor, you construct your :class:``BufferedReader`` as you
    normally would, then construct this object with it as argument.

    Parameters
    ----------
    io : BinaryIO
        IO object used by this class. Depency injection.
    file_size : int
        The fie of the file in bytes.
    callback : Callable[["FileMonitor"], None]
        Callable function used by this class. Depency injection via constructor.

    Attributes
    ----------
    io : BinaryIO
        IO object used by this class. Depency injection.
    callback : Callable[["FileMonitor"], None]
        Callable function used by this class. Depency injection.
    bytes_read : int
      Amount of bytes read from the IO.
    len : int
        Total size of the IO.
    """

    def __init__(self, io: BinaryIO, file_size: int, callback: Callable[["FileMonitor"], None]):
        self.io: BinaryIO = io
        self.callback: Callable[["FileMonitor"], None] = callback

        self.bytes_read: int = 0
        self.len: int = file_size

    def read(self, size: int = -1) -> Any:
        """
        Reads given amount of bytes from configured IO and calls the configured callback for each
        block read. The callback is passed a reference this object that can be used to get current
        self.bytes_read.

        Parameters
        ----------
        size : int, default: -1
            The number of bytes to read. Defaults to -1, so all bytes until EOF are read.

        Returns
        -------
        data: Any
            Data read from the IO.
        """
        data: Any = self.io.read(size)
        self.bytes_read += len(data)
        self.callback(self)

        return data


ByteReadCallback = Callable[[Optional[str], float, float], None]
ProgressCallback = Callable[[int, float], None]
FileUploadCallback = Callable[[str, int, int], None]


class UploadHandler(ABC):
    """
    Holds responsabilities for file upload management and failure into ``RemoteDataset``\\s.

    Parameters
    ----------
    dataset: RemoteDataset
        Target ``RemoteDataset`` where we want to upload our files to.
    local_files : List[LocalFile]
        List of ``LocalFile``\\s to be uploaded.

    Attributes
    ----------
    dataset : RemoteDataset
        Target ``RemoteDataset`` where we want to upload our files to..
    errors : List[UploadRequestError]
        List of errors that happened during the upload process.
    local_files : List[LocalFile]
        List of ``LocalFile``\\s to be uploaded.
    blocked_items : List[ItemPayload]
        List of items that were not able to be uploaded.
    pending_items : List[ItemPayload]
        List of items waiting to be uploaded.
    """

    def __init__(self, dataset: "RemoteDataset", local_files: List[LocalFile]):
        self.dataset: RemoteDataset = dataset
        self.errors: List[UploadRequestError] = []
        self.local_files: List[LocalFile] = local_files
        self._progress: Optional[Iterator[Callable[[Optional[ByteReadCallback]], None]]] = None

        self.blocked_items, self.pending_items = self._request_upload()

    @staticmethod
    def build(dataset: "RemoteDataset", local_files: List[LocalFile]):
        if dataset.version == 1:
            return UploadHandlerV1(dataset, local_files)
        elif dataset.version == 2:
            return UploadHandlerV2(dataset, local_files)
        else:
            raise ValueError(f"Unsupported dataset version: {dataset.version}")

    @property
    def client(self) -> "Client":
        """The ``Client`` used by this ``UploadHander``\\'s ``RemoteDataset``."""
        return self.dataset.client

    @property
    def dataset_identifier(self) -> "DatasetIdentifier":
        """The ``DatasetIdentifier`` of this ``UploadHander``\\'s ``RemoteDataset``."""
        return self.dataset.identifier

    @property
    def blocked_count(self) -> int:
        """Number of items that could not be uploaded successfully."""
        return len(self.blocked_items)

    @property
    def error_count(self) -> int:
        """Number of errors that prevented items from being uploaded."""
        return len(self.errors)

    @property
    def pending_count(self) -> int:
        """Number of items waiting to be uploaded."""
        return len(self.pending_items)

    @property
    def total_count(self) -> int:
        """Total number of blocked and pending items."""
        return self.pending_count + self.blocked_count

    @property
    def progress(self):
        """Current level of upload progress."""
        return self._progress

    def prepare_upload(self) -> Optional[Iterator[Callable[[Optional[ByteReadCallback]], None]]]:
        self._progress = self._upload_files()
        return self._progress

    def upload(
        self,
        multi_threaded: bool = True,
        progress_callback: Optional[ProgressCallback] = None,
        file_upload_callback: Optional[FileUploadCallback] = None,
        max_workers: Optional[int] = None,
    ) -> None:
        if not self._progress:
            self.prepare_upload()

        if progress_callback:
            progress_callback(self.pending_count, 0)

        # needed to ensure that we don't mark a file as completed twice
        file_complete: Set[str] = set()

        def callback(file_name, file_total_bytes, file_bytes_sent):
            if file_upload_callback:
                file_upload_callback(file_name, file_total_bytes, file_bytes_sent)

            if progress_callback:
                if file_total_bytes == file_bytes_sent and file_name not in file_complete:
                    file_complete.add(file_name)
                    progress_callback(self.pending_count, 1)

        if max_workers:
            if max_workers < 1:
                raise ValueError("max_workers must be greater than 0")
            elif max_workers > concurrent.futures.ThreadPoolExecutor()._max_workers:
                raise ValueError(
                    f"max_workers must be less than or equal to {concurrent.futures.ThreadPoolExecutor()._max_workers}"
                )

        if multi_threaded and self.progress:
            with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                future_to_progress = {executor.submit(f, callback) for f in self.progress}
                for future in concurrent.futures.as_completed(future_to_progress):
                    try:
                        future.result()
                    except Exception as exc:
                        print("exception", exc)
        elif self.progress:
            for file_to_upload in self.progress:
                file_to_upload(callback)

    @abstractmethod
    def _request_upload(self) -> Tuple[List[ItemPayload], List[ItemPayload]]:
        pass

    @abstractmethod
    def _upload_files(self) -> Iterator[Callable[[Optional[ByteReadCallback]], None]]:
        pass

    @abstractmethod
    def _upload_file(
        self, dataset_item_id: int, file_path: Path, byte_read_callback: Optional[ByteReadCallback]
    ) -> None:
        pass


class UploadHandlerV1(UploadHandler):
    def __init__(self, dataset: "RemoteDataset", local_files: List[LocalFile]):
        super().__init__(dataset=dataset, local_files=local_files)

    def _request_upload(self) -> Tuple[List[ItemPayload], List[ItemPayload]]:
        blocked_items = []
        items = []
        chunk_size: int = _upload_chunk_size()
        for file_chunk in chunk(self.local_files, chunk_size):
            upload_payload = {"items": [file.data for file in file_chunk]}
            dataset_slug: str = self.dataset_identifier.dataset_slug
            team_slug: Optional[str] = self.dataset_identifier.team_slug

            data: Dict[str, Any] = self.client.upload_data(dataset_slug, upload_payload, team_slug)

            blocked_items.extend([ItemPayload(**item) for item in data["blocked_items"]])
            items.extend([ItemPayload(**item) for item in data["items"]])
        return blocked_items, items

    def _upload_files(self) -> Iterator[Callable[[Optional[ByteReadCallback]], None]]:
        def upload_function(dataset_item_id, local_path) -> Callable[[Optional[ByteReadCallback]], None]:
            return lambda byte_read_callback=None: self._upload_file(dataset_item_id, local_path, byte_read_callback)

        file_lookup = {file.full_path: file for file in self.local_files}
        for item in self.pending_items:
            file = file_lookup.get(item.full_path)
            if not file:
                raise ValueError(f"Cannot match {item.full_path} from payload with files to upload")
            yield upload_function(item.dataset_item_id, file.local_path)

    def _upload_file(
        self, dataset_item_id: int, file_path: Path, byte_read_callback: Optional[ByteReadCallback]
    ) -> None:
        try:
            self._do_upload_file(dataset_item_id, file_path, byte_read_callback)
        except UploadRequestError as e:
            self.errors.append(e)
        except Exception as e:
            self.errors.append(UploadRequestError(file_path=file_path, stage=UploadStage.OTHER, error=e))

    def _do_upload_file(
        self,
        dataset_item_id: int,
        file_path: Path,
        byte_read_callback: Optional[ByteReadCallback] = None,
    ) -> None:
        team_slug: Optional[str] = self.dataset_identifier.team_slug

        try:
            sign_response: Dict[str, Any] = self.client.sign_upload(dataset_item_id, team_slug)
        except Exception as e:
            raise UploadRequestError(file_path=file_path, stage=UploadStage.REQUEST_SIGNATURE, error=e)

        upload_url = sign_response["upload_url"]

        try:
            file_size = file_path.stat().st_size
            if byte_read_callback:
                byte_read_callback(str(file_path), file_size, 0)

            def callback(monitor):
                if byte_read_callback:
                    byte_read_callback(str(file_path), file_size, monitor.bytes_read)

            with file_path.open("rb") as m:
                monitor = FileMonitor(m, file_size, callback)

                retries = 0
                while retries < 5:
                    upload_response = requests.put(f"{upload_url}", data=monitor)
                    # If s3 is getting to many request it will return 503, we will sleep and retry
                    if upload_response.status_code != 503:
                        break

                    time.sleep(2**retries)
                    retries += 1

            upload_response.raise_for_status()
        except Exception as e:
            raise UploadRequestError(file_path=file_path, stage=UploadStage.UPLOAD_TO_S3, error=e)

        try:
            self.client.confirm_upload(dataset_item_id, team_slug)
        except Exception as e:
            raise UploadRequestError(file_path=file_path, stage=UploadStage.CONFIRM_UPLOAD_COMPLETE, error=e)


class UploadHandlerV2(UploadHandler):
    def __init__(self, dataset: "RemoteDataset", local_files: List[LocalFile]):
        super().__init__(dataset=dataset, local_files=local_files)

    def _request_upload(self) -> Tuple[List[ItemPayload], List[ItemPayload]]:
        blocked_items = []
        items = []
        chunk_size: int = _upload_chunk_size()
        for file_chunk in chunk(self.local_files, chunk_size):
            upload_payload = {"items": [file.serialize_v2() for file in file_chunk]}
            dataset_slug: str = self.dataset_identifier.dataset_slug
            team_slug: Optional[str] = self.dataset_identifier.team_slug

            data: Dict[str, Any] = self.client.api_v2.register_data(dataset_slug, upload_payload, team_slug=team_slug)

            blocked_items.extend([ItemPayload.parse_v2(item) for item in data["blocked_items"]])
            items.extend([ItemPayload.parse_v2(item) for item in data["items"]])
        return blocked_items, items

    def _upload_files(self) -> Iterator[Callable[[Optional[ByteReadCallback]], None]]:
        def upload_function(dataset_slug, local_path, upload_id) -> Callable[[Optional[ByteReadCallback]], None]:
            return lambda byte_read_callback=None: self._upload_file(
                dataset_slug, local_path, upload_id, byte_read_callback
            )

        file_lookup = {file.full_path: file for file in self.local_files}
        for item in self.pending_items:
            if len(item.slots) != 1:
                raise NotImplemented("Multi file upload is not supported")
            upload_id = item.slots[0]["upload_id"]
            file = file_lookup.get(item.full_path)
            if not file:
                raise ValueError(f"Cannot match {item.full_path} from payload with files to upload")
            yield upload_function(self.dataset.identifier.dataset_slug, file.local_path, upload_id)

    def _upload_file(
        self, dataset_slug: str, file_path: Path, upload_id: str, byte_read_callback: Optional[ByteReadCallback]
    ) -> None:
        try:
            self._do_upload_file(dataset_slug, file_path, upload_id, byte_read_callback)
        except UploadRequestError as e:
            self.errors.append(e)
        except Exception as e:
            self.errors.append(UploadRequestError(file_path=file_path, stage=UploadStage.OTHER, error=e))

    def _do_upload_file(
        self,
        dataset_slug: str,
        file_path: Path,
        upload_id: str,
        byte_read_callback: Optional[ByteReadCallback] = None,
    ) -> None:
        team_slug: Optional[str] = self.dataset_identifier.team_slug

        try:
            sign_response: Dict[str, Any] = self.client.api_v2.sign_upload(dataset_slug, upload_id, team_slug=team_slug)
        except Exception as e:
            raise UploadRequestError(file_path=file_path, stage=UploadStage.REQUEST_SIGNATURE, error=e)

        upload_url = sign_response["upload_url"]

        try:
            file_size = file_path.stat().st_size
            if byte_read_callback:
                byte_read_callback(str(file_path), file_size, 0)

            def callback(monitor):
                if byte_read_callback:
                    byte_read_callback(str(file_path), file_size, monitor.bytes_read)

            with file_path.open("rb") as m:
                monitor = FileMonitor(m, file_size, callback)

                retries = 0
                while retries < 5:
                    upload_response = requests.put(f"{upload_url}", data=monitor)
                    # If s3 is getting to many request it will return 503, we will sleep and retry
                    if upload_response.status_code != 503:
                        break

                    time.sleep(2**retries)
                    retries += 1

            upload_response.raise_for_status()
        except Exception as e:
            raise UploadRequestError(file_path=file_path, stage=UploadStage.UPLOAD_TO_S3, error=e)

        try:
            self.client.api_v2.confirm_upload(dataset_slug, upload_id, team_slug=team_slug)
        except Exception as e:
            raise UploadRequestError(file_path=file_path, stage=UploadStage.CONFIRM_UPLOAD_COMPLETE, error=e)


DEFAULT_UPLOAD_CHUNK_SIZE: int = 500


def _upload_chunk_size() -> int:
    """
    Gets the chunk size to be used from the OS environment, or uses the default one if that is not
    possible. The default chunk size is 500.

    Returns
    -------
    int
        The chunk size to be used.
    """
    env_chunk: Optional[str] = os.getenv("DARWIN_UPLOAD_CHUNK_SIZE")
    if env_chunk is None:
        return DEFAULT_UPLOAD_CHUNK_SIZE

    try:
        return int(env_chunk)
    except ValueError:
        print("Cannot cast environment variable DEFAULT_UPLOAD_CHUNK_SIZE to integer")
        print(f"Setting chunk size to {DEFAULT_UPLOAD_CHUNK_SIZE}")
        return DEFAULT_UPLOAD_CHUNK_SIZE



from enum import Enum


class DocEnum(Enum):
    """
    Documenting Enums in Python is not supported by many tools. Therefore this class was created to
    support just that. It is basically a hack to allow Enum documentation.
    `See more here <https://stackoverflow.com/a/50473952/1337392>`
    """

    def __new__(cls, value, doc=None):
        self = object.__new__(cls)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self



from typing import List, Optional

from pydantic import PositiveInt, validator

from darwin.future.data_objects.release import ReleaseList
from darwin.future.data_objects.validators import parse_name
from darwin.future.pydantic_base import DefaultDarwin


class Dataset(DefaultDarwin):
    """A class to manage all the information around a dataset on the darwin platform, including validation

    Attributes
    ----------
    name : str
    slug : str
    id: Optional[int] = None
    releases: Optional[List[Release]] = None
        - a list of export releases linked to a dataset
    Methods
    ----------
    _name_validator: validates and auto formats the name variable
    """

    name: str
    slug: str
    id: Optional[PositiveInt] = None
    releases: Optional[ReleaseList] = None

    # Data Validation
    _name_validator = validator("name", allow_reuse=True)(parse_name)


DatasetList = List[Dataset]



import json as native_json
from asyncore import loop
from pathlib import Path
from typing import Iterable
from rich.console import Console

console = Console()
try:
    import nibabel as nib
    from nibabel.orientations import io_orientation, axcodes2ornt, ornt_transform
except ImportError:
    import_fail_string = """
    You must install darwin-py with pip install darwin-py\[medical]
    in order to export using nifti format
    """
    console.print(import_fail_string)
    exit()
import numpy as np
import orjson as json
from PIL import Image

import darwin.datatypes as dt
from darwin.utils import convert_polygons_to_mask, get_progress_bar


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into nifti format inside of the given
    ``output_dir``. Deletes everything within ``output_dir/masks`` before writting to it.

    Parameters
    ----------
    annotation_files : Iterable[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new instance mask files will be.
    """
    video_annotations = list(annotation_files)
    for video_annotation in video_annotations:
        export_single_nifti_file(video_annotation, output_dir)


def export_single_nifti_file(video_annotation: dt.AnnotationFile, output_dir: Path) -> None:
    output_volumes = None
    filename = Path(video_annotation.filename)
    suffixes = filename.suffixes
    if len(suffixes) > 2:
        return create_error_message_json(
            "Misconfigured filename, contains too many suffixes", output_dir, str(filename)
        )
    elif len(suffixes) == 2:
        if suffixes[0] == ".nii" and suffixes[1] == ".gz":
            image_id = str(filename).strip("".join(suffixes))
        else:
            return create_error_message_json("Two suffixes found but not ending in .nii.gz", output_dir, str(filename))
    elif len(suffixes) == 1:
        if suffixes[0] == ".nii" or suffixes[0] == ".dcm":
            image_id = filename.stem
        else:
            return create_error_message_json(
                "Misconfigured filename, not ending in .nii or .dcm. Are you sure this is medical data?",
                output_dir,
                str(filename),
            )
    else:
        return create_error_message_json(
            "You are trying to export to nifti. Filename should contain either .nii, .nii.gz or .dcm extension."
            "Are you sure this is medical data?",
            output_dir,
            str(filename),
        )
    if video_annotation is None:
        return create_error_message_json("video_annotation not found", output_dir, image_id)
    # Pick the first slot to take the metadata from. We assume that all slots have the same metadata.
    metadata = video_annotation.slots[0].metadata
    if metadata is None:
        return create_error_message_json(
            f"No metadata found for {str(filename)}, are you sure this is medical data?", output_dir, image_id
        )
    volume_dims, pixdim, affine, original_affine = process_metadata(metadata)
    if affine is None or pixdim is None or volume_dims is None:
        return create_error_message_json(
            f"Missing one of affine, pixdim or shape in metadata for {str(filename)}, try reuploading file",
            output_dir,
            image_id,
        )
    if not video_annotation.annotations:
        create_empty_nifti_file(volume_dims, affine, output_dir, image_id)
    # Builds a map of class to integer
    class_map = {}
    class_count = 1
    for _, annotation in enumerate(video_annotation.annotations):
        frames = annotation.frames
        for frame_idx in frames.keys():
            class_name = frames[frame_idx].annotation_class.name
            if class_name not in class_map:
                class_map[class_name] = class_count
                class_count += 1
    # Builds output volumes per class
    if output_volumes is None:
        output_volumes = {class_name: np.zeros(volume_dims) for class_name in class_map.keys()}
    # Loops through annotations to build volumes
    for _, annotation in enumerate(video_annotation.annotations):
        frames = annotation.frames
        for frame_idx in frames.keys():
            view_idx = get_view_idx_from_slot_name(annotation.slot_names[0])
            if view_idx == 0:
                height, width = volume_dims[0], volume_dims[1]
                pixdims = [pixdim[0], pixdim[1]]
            elif view_idx == 1:
                height, width = volume_dims[0], volume_dims[2]
                pixdims = [pixdim[0], pixdim[2]]
            elif view_idx == 2:
                height, width = volume_dims[1], volume_dims[2]
                pixdims = [pixdim[1], pixdim[2]]
            if "paths" in frames[frame_idx].data:
                # Dealing with a complex polygon
                polygons = [shift_polygon_coords(polygon_path) for polygon_path in frames[frame_idx].data["paths"]]
            elif "path" in frames[frame_idx].data:
                # Dealing with a simple polygon
                polygons = shift_polygon_coords(frames[frame_idx].data["path"])
            else:
                continue
            class_name = frames[frame_idx].annotation_class.name
            im_mask = convert_polygons_to_mask(polygons, height=height, width=width)
            output_volume = output_volumes[class_name]
            if view_idx == 0:
                output_volume[:, :, frame_idx] = np.logical_or(im_mask, output_volume[:, :, frame_idx])
            elif view_idx == 1:
                output_volume[:, frame_idx, :] = np.logical_or(im_mask, output_volume[:, frame_idx, :])
            elif view_idx == 2:
                output_volume[frame_idx, :, :] = np.logical_or(im_mask, output_volume[frame_idx, :, :])
    for class_name in class_map.keys():
        img = nib.Nifti1Image(
            dataobj=np.flip(output_volumes[class_name], (0, 1, 2)).astype(np.int16),
            affine=affine,
        )
        if original_affine is not None:
            orig_ornt = io_orientation(original_affine)  # Get orientation of current affine
            img_ornt = io_orientation(affine) # Get orientation of RAS affine
            from_canonical = ornt_transform(img_ornt, orig_ornt)  # Get transform from RAS to current affine
            img = img.as_reoriented(from_canonical)
        output_path = Path(output_dir) / f"{image_id}_{class_name}.nii.gz"
        if not output_path.parent.exists():
            output_path.parent.mkdir(parents=True)
        nib.save(img=img, filename=output_path)


def shift_polygon_coords(polygon):
    # Need to make it clear that we flip x/y because we need to take the transpose later.
    # TODO: We might need to add a correction factor here for anisotropic pixdims.
    return [{"x": p["y"], "y": p["x"]} for p in polygon]


def get_view_idx(frame_idx, groups):
    if groups is None:
        return 0
    for view_idx, group in enumerate(groups):
        if frame_idx in group:
            return view_idx


def get_view_idx_from_slot_name(slot_name):
    # if mpr:
    #     #do this correct treatment volumetrically.
    #     pass
    slot_names = {"0.1": 0, "0.2": 1, "0.3": 2}
    slot_names.get(slot_name, 0)
    return slot_names.get(slot_name, 0)


def process_metadata(metadata):
    volume_dims = metadata.get("shape")
    pixdim = metadata.get("pixdim")
    affine = process_affine(metadata.get("affine"))
    original_affine = process_affine(metadata.get("original_affine"))
    # If the original affine is in the medical payload of metadata then use it
    if isinstance(pixdim, str):
        pixdim = eval(pixdim)
        if isinstance(pixdim, tuple) or isinstance(pixdim, list):
            if len(pixdim) == 4:
                pixdim = pixdim[1:]
            if len(pixdim) != 3:
                pixdim = None
        else:
            pixdim = None
    if isinstance(volume_dims, list):
        if volume_dims:
            if volume_dims[0] == 1:  # remove first singleton dimension
                volume_dims = volume_dims[1:]
        else:
            volume_dims = None
    return volume_dims, pixdim, affine, original_affine


def process_affine(affine):
    if isinstance(affine, str):
        affine = np.squeeze(np.array([eval(l) for l in affine.split("\n")]))
    elif isinstance(affine, list):
        affine = np.array(affine).astype(np.float)
    else:
        return
    if isinstance(affine, np.ndarray):
        return affine


def create_error_message_json(error_message, output_dir, image_id: str):
    output_path = Path(output_dir) / f"{image_id}_error.json"
    if not output_path.parent.exists():
        output_path.parent.mkdir(parents=True)
    with open(output_path, "w") as f:
        native_json.dump({"error": error_message}, f)


def create_empty_nifti_file(volume_dims, affine, output_dir, image_id: str):
    output_path = Path(output_dir) / f"{image_id}_empty.nii.gz"
    if not output_path.parent.exists():
        output_path.parent.mkdir(parents=True)
    img = nib.Nifti1Image(dataobj=np.flip(np.zeros(volume_dims), (0, 1, 2)).astype(np.int16), affine=affine)
    nib.save(img=img, filename=output_path)



import unittest
from typing import List

import pytest

from darwin.future.core.types import query as Query
from darwin.future.data_objects.team import Team


@pytest.fixture
def basic_filters() -> List[Query.QueryFilter]:
    return [
        Query.QueryFilter(name="test1", param="test1"),
        Query.QueryFilter(name="test2", param="test2"),
        Query.QueryFilter(name="test3", param="test3"),
    ]


@pytest.fixture
def test_team() -> Team:
    return Team(slug="test-team", id=0)


def test_query_instantiated(basic_filters: List[Query.QueryFilter], test_team: Team) -> None:
    q = Query.Query(test_team, basic_filters)
    assert q.filters == basic_filters


def test_query_filter_functionality(basic_filters: List[Query.QueryFilter], test_team: Team) -> None:
    q = Query.Query(test_team)
    for f in basic_filters:
        q = q.filter(f)
    assert q.filters == basic_filters
    dropped_filter = basic_filters.pop(0)
    q = q - dropped_filter
    assert q.filters == basic_filters
    basic_filters.append(dropped_filter)
    q = q + dropped_filter
    assert q.filters == basic_filters

    # Test filter drops from middle
    dropped_filter = basic_filters.pop(1)
    q = q - dropped_filter
    assert q.filters == basic_filters
    basic_filters.append(dropped_filter)
    q = q + dropped_filter
    assert q.filters == basic_filters


def test_query_iterable(basic_filters: List[Query.QueryFilter], test_team: Team) -> None:
    q = Query.Query(test_team, basic_filters)
    for i, f in enumerate(q):
        assert f == basic_filters[i]
    assert q.filters is not None
    assert len(q) == len(basic_filters)



__all__ = ["main"]

import getpass
import os
import platform
from argparse import ArgumentParser, Namespace
from datetime import datetime
from json import dumps

import requests.exceptions

import darwin.cli_functions as f
from darwin import __version__
from darwin.exceptions import InvalidTeam, Unauthenticated, Unauthorized
from darwin.options import Options


def main() -> None:
    """
    Executes the main function of program.

    Raises
    ------
    Unauthorized
        If the API key with which the use is authenticated does not grant access for the given
        action.
    Unauthenticated
        If a given action needs authentication and you are not authenticated.
    InvalidTeam
        If you are trying to use a team that is not specified in the configuration file. To fix this
        please authenticate with the given team first.
    requests.exceptions.ConnectionError
        If there is a connection issue.
    """
    args, parser = Options().parse_args()
    try:
        _run(args, parser)
    except Unauthorized:
        f._error("Your API key is not authorized to do that action.")
    except Unauthenticated:
        f._error("You need to specify a valid API key to do that action.")
    except InvalidTeam:
        f._error("The team specified is not in the configuration, please authenticate first.")
    except requests.exceptions.ConnectionError:
        f._error("Darwin seems unreachable, please try again in a minute or contact support.")
    except Exception as e:  # Catch unhandled exceptions
        filename = f"darwin_error_{datetime.now().timestamp()}.log"

        fd = open(filename, "w")
        fd.write("Darwin CLI error log")
        fd.write(f"Version: {__version__}")
        fd.write(f"OS: {platform.platform()}")
        fd.write(f"Command: {str(args)}")
        fd.write(f"Error: {str(e)}")
        fd.close()

        f._error(
            f"An unexpected error occurred, errors have been written to {filename}, please contact support, and send them the file."
            + f"Error: {str(e)}"
        )


def _run(args: Namespace, parser: ArgumentParser) -> None:
    if args.command == "help":
        f.help(parser)

    # Authenticate user
    if args.command == "authenticate":
        api_key = os.getenv("DARWIN_API_KEY")
        if api_key:
            print("Using API key from DARWIN_API_KEY")
        else:
            api_key = getpass.getpass(prompt="API key: ", stream=None)
            api_key = api_key.strip()
            if api_key == "":
                print("API Key needed, generate one for your team: https://darwin.v7labs.com/?settings=api-keys")
                return
        f.authenticate(api_key)
        print("Authentication succeeded.")
    # Select / List team
    elif args.command == "team":
        if args.team_name:
            f.set_team(args.team_name)
        elif args.current:
            f.current_team()
        else:
            f.list_teams()
    # Set compression level
    elif args.command == "compression":
        f.set_compression_level(args.compression_level)
    # Version
    elif args.command == "version":
        print(__version__)

    elif args.command == "convert":
        f.convert(args.format, args.files, args.output_dir)
    elif args.command == "dataset":
        if args.action == "remote":
            f.list_remote_datasets(args.all, args.team)
        elif args.action == "local":
            f.local(args.team)
        elif args.action == "create":
            f.create_dataset(args.dataset)
        elif args.action == "path":
            path = f.path(args.dataset)
            if path:
                print(path)
        # Print the url of a remote project
        elif args.action == "url":
            f.url(args.dataset)
        elif args.action == "push":
            f.upload_data(
                args.dataset,
                args.files,
                args.exclude,
                args.fps,
                args.path,
                args.frames,
                args.extract_views,
                args.preserve_folders,
                args.verbose,
            )
        # Remove a project (remotely)
        elif args.action == "remove":
            f.remove_remote_dataset(args.dataset)
        elif args.action == "report":
            f.dataset_report(args.dataset, args.granularity or "day", args.pretty)
        elif args.action == "export":
            f.export_dataset(
                args.dataset, args.include_url_token, args.name, args.class_ids, args.include_authorship, args.version
            )
        elif args.action == "files":
            f.list_files(args.dataset, args.status, args.path, args.only_filenames, args.sort_by)
        elif args.action == "releases":
            f.dataset_list_releases(args.dataset)
        elif args.action == "pull":
            f.pull_dataset(args.dataset, args.only_annotations, args.folders, args.video_frames, args.force_slots)
        elif args.action == "import":
            f.dataset_import(
                args.dataset,
                args.format,
                args.files,
                args.append,
                not args.yes,
                args.delete_for_empty,
                args.import_annotators,
                args.import_reviewers,
                cpu_limit=args.cpu_limit,
            ),
        elif args.action == "convert":
            f.dataset_convert(args.dataset, args.format, args.output_dir)
        elif args.action == "set-file-status":
            f.set_file_status(args.dataset, args.status, args.files)
        elif args.action == "delete-files":
            f.delete_files(args.dataset, args.files, args.yes)
        elif args.action == "split":
            f.split(args.dataset, args.val_percentage, args.test_percentage, args.seed)
        elif args.action == "help" or args.action is None:
            f.help(parser, "dataset")
        elif args.action == "comment":
            f.post_comment(
                args.dataset,
                args.file,
                args.text,
                args.x,
                args.y,
                args.w,
                args.h,
            )
    # Annotation schema validation
    elif args.command == "validate":
        f.validate_schemas(location=args.location, pattern=args.pattern, silent=args.silent, output=args.output)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Any, Dict, Iterable
from xml.etree.ElementTree import Element, SubElement, tostring

import deprecation
import orjson as json

import darwin.datatypes as dt
from darwin.utils import attempt_decode
from darwin.version import __version__

DEPRECATION_MESSAGE = """

This function is going to be turned into private. This means that breaking 
changes in its interface and implementation are to be expected. We encourage using ``export`` 
instead of calling this low-level function directly.

"""

REMOVAL_MESSAGE = """

This function is going to be removed. This means that breaking 
changes in its interface and implementation are to be expected. We encourage no longer using it.

"""


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into the pascalvoc format inside of the given
    ``output_dir``.

    Parameters
    ----------
    annotation_files : Iterable[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new pascalvoc files will be.
    """
    for annotation_file in annotation_files:
        _export_file(annotation_file, output_dir)


@deprecation.deprecated(
    deprecated_in="0.7.10",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def export_file(annotation_file: dt.AnnotationFile, output_dir: Path) -> None:
    xml = build_xml(annotation_file)
    output_file_path = (output_dir / annotation_file.filename).with_suffix(".xml")
    output_file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_file_path, "wb") as f:
        f.write(tostring(xml))


@deprecation.deprecated(
    deprecated_in="0.7.10",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_xml(annotation_file: dt.AnnotationFile) -> Element:
    root: Element = Element("annotation")
    add_subelement_text(root, "folder", "images")
    add_subelement_text(root, "filename", annotation_file.filename)
    add_subelement_text(root, "path", f"images/{annotation_file.filename}")

    source = SubElement(root, "source")
    add_subelement_text(source, "database", "darwin")

    size = SubElement(root, "size")
    add_subelement_text(size, "width", str(annotation_file.image_width))
    add_subelement_text(size, "height", str(annotation_file.image_height))
    add_subelement_text(size, "depth", "3")

    add_subelement_text(root, "segmented", "0")

    for annotation in annotation_file.annotations:
        annotation_type = annotation.annotation_class.annotation_type
        if annotation_type not in ["bounding_box", "polygon", "complex_polygon"]:
            continue

        data = annotation.data
        sub_annotation = SubElement(root, "object")
        add_subelement_text(sub_annotation, "name", annotation.annotation_class.name)
        add_subelement_text(sub_annotation, "pose", "Unspecified")
        add_subelement_text(sub_annotation, "truncated", "0")
        add_subelement_text(sub_annotation, "difficult", "0")
        bndbox = SubElement(sub_annotation, "bndbox")

        if annotation_type == "polygon" or annotation_type == "complex_polygon":
            data = data.get("bounding_box")

        xmin = data.get("x")
        ymin = data.get("y")
        xmax = xmin + data.get("w")
        ymax = ymin + data.get("h")
        add_subelement_text(bndbox, "xmin", str(round(xmin)))
        add_subelement_text(bndbox, "ymin", str(round(ymin)))
        add_subelement_text(bndbox, "xmax", str(round(xmax)))
        add_subelement_text(bndbox, "ymax", str(round(ymax)))

    return root


@deprecation.deprecated(
    deprecated_in="0.7.10",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def add_subelement_text(parent: Element, name: str, value: Any) -> Element:
    sub: Element = SubElement(parent, name)
    sub.text = value
    return sub


@deprecation.deprecated(
    deprecated_in="0.7.10",
    removed_in="0.8.0",
    current_version=__version__,
    details=REMOVAL_MESSAGE,
)
def convert_file(path: Path) -> Element:
    data = attempt_decode(path)
    return build_voc(data["image"], data["annotations"])


@deprecation.deprecated(
    deprecated_in="0.7.10",
    removed_in="0.8.0",
    current_version=__version__,
    details=REMOVAL_MESSAGE,
)
def save_xml(xml: Element, path: Path) -> None:
    with open(path, "wb") as f:
        f.write(tostring(xml))


@deprecation.deprecated(
    deprecated_in="0.7.10",
    removed_in="0.8.0",
    current_version=__version__,
    details=REMOVAL_MESSAGE,
)
def build_voc(metadata: Dict[str, Any], annotations: Iterable[Dict[str, Any]]) -> Element:
    print(metadata)
    root: Element = Element("annotation")
    add_subelement_text(root, "folder", "images")
    add_subelement_text(root, "filename", metadata["original_filename"])
    add_subelement_text(root, "path", f"images/{metadata['original_filename']}")

    source: Element = SubElement(root, "source")
    add_subelement_text(source, "database", "darwin")

    size: Element = SubElement(root, "size")
    add_subelement_text(size, "width", str(metadata["width"]))
    add_subelement_text(size, "height", str(metadata["height"]))
    add_subelement_text(size, "depth", "3")

    add_subelement_text(root, "segmented", "0")

    for annotation in annotations:
        if "bounding_box" not in annotation:
            continue
        data = annotation["bounding_box"]
        sub_annotation = SubElement(root, "object")
        add_subelement_text(sub_annotation, "name", annotation["name"])
        add_subelement_text(sub_annotation, "pose", "Unspecified")
        add_subelement_text(sub_annotation, "truncated", "0")
        add_subelement_text(sub_annotation, "difficult", "0")
        bndbox = SubElement(sub_annotation, "bndbox")
        add_subelement_text(bndbox, "xmin", str(round(data["x"])))
        add_subelement_text(bndbox, "ymin", str(round(data["y"])))
        add_subelement_text(bndbox, "xmax", str(round(data["x"] + data["w"])))
        add_subelement_text(bndbox, "ymax", str(round(data["y"] + data["h"])))

    return root


######################################


def _export_file(annotation_file: dt.AnnotationFile, output_dir: Path) -> None:
    xml = _build_xml(annotation_file)
    output_file_path = (output_dir / annotation_file.filename).with_suffix(".xml")
    output_file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_file_path, "wb") as f:
        f.write(tostring(xml))


def _build_xml(annotation_file: dt.AnnotationFile) -> Element:
    root: Element = Element("annotation")
    _add_subelement_text(root, "folder", "images")
    _add_subelement_text(root, "filename", annotation_file.filename)
    _add_subelement_text(root, "path", f"images/{annotation_file.filename}")

    source = SubElement(root, "source")
    _add_subelement_text(source, "database", "darwin")

    size = SubElement(root, "size")
    _add_subelement_text(size, "width", str(annotation_file.image_width))
    _add_subelement_text(size, "height", str(annotation_file.image_height))
    _add_subelement_text(size, "depth", "3")

    _add_subelement_text(root, "segmented", "0")

    for annotation in annotation_file.annotations:
        annotation_type = annotation.annotation_class.annotation_type
        if annotation_type not in ["bounding_box", "polygon", "complex_polygon"]:
            continue

        data = annotation.data
        sub_annotation = SubElement(root, "object")
        _add_subelement_text(sub_annotation, "name", annotation.annotation_class.name)
        _add_subelement_text(sub_annotation, "pose", "Unspecified")
        _add_subelement_text(sub_annotation, "truncated", "0")
        _add_subelement_text(sub_annotation, "difficult", "0")
        bndbox = SubElement(sub_annotation, "bndbox")

        if annotation_type == "polygon" or annotation_type == "complex_polygon":
            data = data.get("bounding_box")

        xmin = data.get("x")
        ymin = data.get("y")
        xmax = xmin + data.get("w")
        ymax = ymin + data.get("h")
        _add_subelement_text(bndbox, "xmin", str(round(xmin)))
        _add_subelement_text(bndbox, "ymin", str(round(ymin)))
        _add_subelement_text(bndbox, "xmax", str(round(xmax)))
        _add_subelement_text(bndbox, "ymax", str(round(ymax)))

    return root


def _add_subelement_text(parent: Element, name: str, value: Any) -> Element:
    sub: Element = SubElement(parent, name)
    sub.text = value
    return sub



import pytest

from darwin.future.core.client import Client, DarwinConfig
from darwin.future.data_objects.team import Team, TeamMember


@pytest.fixture
def base_config() -> DarwinConfig:
    return DarwinConfig(
        api_key="test_key",
        base_url="http://test_url.com/",
        api_endpoint="http://test_url.com/api/",
        default_team="default-team",
        teams={},
    )


@pytest.fixture
def base_client(base_config: DarwinConfig) -> Client:
    return Client(base_config)


@pytest.fixture
def base_team_json() -> dict:
    return {"slug": "test-team", "id": "0"}


@pytest.fixture
def base_team(base_team_json: dict) -> Team:
    return Team.parse_obj(base_team_json)


@pytest.fixture
def base_team_member_json() -> dict:
    return {
        "email": "email",
        "id": "0",
        "first_name": "first",
        "last_name": "last",
        "role": "member",
        "team_id": "0",
        "user_id": "0",
    }


@pytest.fixture
def base_team_member(base_team_member_json: dict) -> TeamMember:
    return TeamMember.parse_obj(base_team_member_json)



"""
Holds functions that convert backend errors into a pythonic format the application can understand.
"""

from typing import Any, Dict

from darwin.exceptions import NameTaken, ValidationError


def name_taken(code: int, body: Dict[str, Any]) -> None:
    """
    Validates if a request to the backend errored out with a NameTaken error.

    Parameters
    ----------
    code : int
        The response code.
    body : Dict[str, Any]
        The response body.

    Raises
    ------
    NameTaken
        If both ``code`` and ``body`` indicate that the server request errored due to a name being
        already taken.
    """
    if code != 422:
        return
    if body.get("errors", {}).get("name") == ["has already been taken"]:
        raise NameTaken


def validation_error(code: int, body: Dict[str, Any]) -> None:
    """
    Validates if a request to the backend errored out with a Validation error.

    Parameters
    ----------
    code : int
        The response code.
    body : Dict[str, Any]
        The response body.

    Raises
    ------
    ValidationError
        If both ``code`` and ``body`` indicate that the server request errored because it failed
        validation.
    """
    if code == 422:
        raise ValidationError(body)



import xml.etree.ElementTree as ET
from pathlib import Path
from typing import List, Optional

import darwin.datatypes as dt


def parse_path(path: Path) -> Optional[dt.AnnotationFile]:
    """
    Parses the given pascalvoc file and maybe returns the corresponding annotation.
    The file must have the following structure:

    .. code-block:: xml

        <filename>SOME_FILE_NAME</filename>
        <object>
            <name>CLASS_NAME</name>
            <bndbox>
                <xmax>NUMBER</xmax>
                <xmin>NUMBER</xmin>
                <ymax>NUMBER</ymax>
                <ymin>NUMBER</ymin>
            </bndbox>
        </object>
        <object>
            ...
        </object>

    Parameters
    --------
    path: Path
        The path of the file to parse.

    Returns
    -------
    Optional[darwin.datatypes.AnnotationFile]
        An AnnotationFile with the parsed information from the file or None, if the file is not a
        `XML` file.

    Raises
    ------
    ValueError
        If a mandatory child element is missing or is empty. Mandatory child elements are:
        filename, name, bndbox, xmin, xmax, ymin and ymax.

    """
    if path.suffix != ".xml":
        return None

    tree = ET.parse(str(path))
    root = tree.getroot()

    filename = _find_text_value(root, "filename")

    annotations: List[dt.Annotation] = list(filter(None, map(_parse_annotation, root.findall("object"))))
    annotation_classes = set([annotation.annotation_class for annotation in annotations])

    return dt.AnnotationFile(path, filename, annotation_classes, annotations, remote_path="/")


def _parse_annotation(annotation_object: ET.Element) -> dt.Annotation:
    """
    Parses the given XML element and returns the corresponding annotation.

    Parameters
    --------
    annotation_object: xml.etree.ElementTree.Element
        The element to convert into an annotation.

    Returns
    -------
    darwin.datatypes.AnnotationFile
        An AnnotationFile with the parsed information from the XML element.

    Raises
    ------
    ValueError
        If a mandatory chield element is missing or is empty. Mandatory child elements are:
        name, bndbox, xmin, xmax, ymin and ymax.
    """
    class_name = _find_text_value(annotation_object, "name")

    bndbox = _find_element(annotation_object, "bndbox")
    xmin = int(float(_find_text_value(bndbox, "xmin")))
    xmax = int(float(_find_text_value(bndbox, "xmax")))
    ymin = int(float(_find_text_value(bndbox, "ymin")))
    ymax = int(float(_find_text_value(bndbox, "ymax")))

    return dt.make_bounding_box(class_name, xmin, ymin, xmax - xmin, ymax - ymin)


def _find_element(source: ET.Element, name: str) -> ET.Element:
    """
    Finds a child element inside the source element with the given name and returns it.

    Parameters
    --------
    source: xml.etree.ElementTree.Element
        Parent element that contains childs elements to be searched.
    name: str
        Name of the child element we wish to find.

    Returns
    -------
    xml.etree.ElementTree.Element
        Child element with the given name.

    Raises
    ------
    ValueError
        If a child element with the given name could not be found.
    """
    element = source.find(name)
    if element is None:
        raise ValueError(f"Could not find {name} element in annotation file")
    return element


def _find_text_value(source: ET.Element, name: str) -> str:
    """
    Finds a child element inside the source element with the given name and returns its text value.

    Parameters
    --------
    source: xml.etree.ElementTree.Element
        Parent element that contains childs elements to be searched.
    name: str
        Name of the child element we wish to find.

    Returns
    -------
    str
        Text value of the found child element.

    Raises
    ------
    ValueError
        If the found child element has no text value or its text value is empty.
    """
    element = _find_element(source, name)
    if element.text is None or not element.text.strip():
        raise ValueError(f"{name} element does not have a text value")
    return element.text



from pathlib import Path
from typing import Any, Dict, Iterable, List, Union

import orjson as json

from darwin.datatypes import (
    Annotation,
    AnnotationAuthor,
    AnnotationClass,
    AnnotationFile,
    DictFreeForm,
    SubAnnotation,
    VideoAnnotation,
)
from darwin.exceptions import (
    ExportException,
    ExportException_CouldNotAssembleOutputPath,
    ExportException_CouldNotBuildOutput,
    ExportException_CouldNotWriteFile,
)


def export(annotation_files: Iterable[AnnotationFile], output_dir: Path) -> None:
    errors: List[Exception] = []

    for id, annotation_file in enumerate(annotation_files):
        try:
            _export_file(annotation_file, id, output_dir)
        except Exception as e:
            errors.append(e)

    if errors:
        raise ExportException.from_multiple_exceptions(errors)


def _export_file(annotation_file: AnnotationFile, _: int, output_dir: Path) -> None:

    try:
        filename = annotation_file.path.parts[-1]
        output_file_path = (output_dir / filename).with_suffix(".json")
    except Exception as e:
        raise ExportException_CouldNotAssembleOutputPath(
            f"Could not export file {annotation_file.path} to {output_dir}"
        ) from e

    try:
        output: DictFreeForm = _build_json(annotation_file)
    except Exception as e:
        raise ExportException_CouldNotBuildOutput(f"Could not build output for {annotation_file.path}") from e

    try:
        with open(output_file_path, "w") as f:
            op = json.dumps(output, option=json.OPT_INDENT_2 | json.OPT_SERIALIZE_NUMPY | json.OPT_NON_STR_KEYS).decode(
                "utf-8"
            )
            f.write(op)
    except Exception as e:
        raise ExportException_CouldNotWriteFile(f"Could not write output for {annotation_file.path}") from e


def _build_json(annotation_file: AnnotationFile) -> DictFreeForm:
    if annotation_file.is_video:
        return _build_video_json(annotation_file)
    else:
        return _build_image_json(annotation_file)


def _build_image_json(annotation_file: AnnotationFile) -> DictFreeForm:
    return {
        "image": {
            "seq": annotation_file.seq,
            "width": annotation_file.image_width,
            "height": annotation_file.image_height,
            "filename": annotation_file.filename,
            "original_filename": annotation_file.filename,
            "url": annotation_file.image_url,
            "thumbnail_url": annotation_file.image_thumbnail_url,
            "path": annotation_file.remote_path,
            "workview_url": annotation_file.workview_url,
            **_build_metadata(annotation_file),
        },
        "annotations": list(map(_build_annotation, annotation_file.annotations)),
        "dataset": str(annotation_file.dataset_name),
    }


def _build_video_json(annotation_file: AnnotationFile) -> DictFreeForm:
    return {
        "image": {
            "seq": annotation_file.seq,
            "frame_urls": annotation_file.frame_urls,
            "frame_count": len(annotation_file.frame_urls or []),
            "width": annotation_file.image_width,
            "height": annotation_file.image_height,
            "filename": annotation_file.filename,
            "original_filename": annotation_file.filename,
            "thumbnail_url": annotation_file.image_thumbnail_url,
            "url": annotation_file.image_url,
            "path": annotation_file.remote_path,
            "workview_url": annotation_file.workview_url,
            **_build_metadata(annotation_file),
        },
        "annotations": list(map(_build_annotation, annotation_file.annotations)),
        "dataset": str(annotation_file.dataset_name),
    }


def _build_annotation(annotation: Union[Annotation, VideoAnnotation]) -> DictFreeForm:
    if isinstance(annotation, VideoAnnotation):
        return _build_video_annotation(annotation)
    else:
        return _build_image_annotation(annotation)


def _build_author(author: AnnotationAuthor) -> DictFreeForm:
    return {"full_name": author.name, "email": author.email}


def _build_sub_annotation(sub: SubAnnotation) -> DictFreeForm:
    if sub.annotation_type == "instance_id":
        return {sub.annotation_type: {"value": sub.data}}
    elif sub.annotation_type == "attributes":
        return {sub.annotation_type: sub.data}
    elif sub.annotation_type == "text":
        return {sub.annotation_type: {"text": sub.data}}
    else:
        return {sub.annotation_type: sub.data}


def _build_authorship(annotation: Union[VideoAnnotation, Annotation]) -> DictFreeForm:
    annotators = {}
    if annotation.annotators:
        annotators = {"annotators": [_build_author(annotator) for annotator in annotation.annotators]}

    reviewers = {}
    if annotation.reviewers:
        reviewers = {"annotators": [_build_author(reviewer) for reviewer in annotation.reviewers]}

    return {**annotators, **reviewers}


def _build_video_annotation(annotation: VideoAnnotation) -> DictFreeForm:
    return {
        **annotation.get_data(
            only_keyframes=False,
            post_processing=lambda annotation, _: _build_image_annotation(annotation, skip_slots=True),
        ),
        "name": annotation.annotation_class.name,
        "slot_names": annotation.slot_names,
        **_build_authorship(annotation),
    }


def _build_image_annotation(annotation: Annotation, skip_slots: bool = False) -> DictFreeForm:
    json_subs = {}
    for sub in annotation.subs:
        json_subs.update(_build_sub_annotation(sub))

    base_json = {
        **json_subs,
        **_build_authorship(annotation),
        **_build_legacy_annotation_data(annotation.annotation_class, annotation.data),
        "name": annotation.annotation_class.name,
    }

    if skip_slots:
        return base_json
    else:
        return {**base_json, "slot_names": annotation.slot_names}


def _build_legacy_annotation_data(annotation_class: AnnotationClass, data: DictFreeForm) -> DictFreeForm:
    if annotation_class.annotation_type == "complex_polygon":
        data["path"] = data["paths"]
        del data["paths"]
        return {"complex_polygon": data}
    else:
        return {annotation_class.annotation_type: data}


def _build_metadata(annotation_file: AnnotationFile) -> DictFreeForm:
    if len(annotation_file.slots) > 0 and annotation_file.slots[0].metadata:
        return {"metadata": annotation_file.slots[0].metadata}
    else:
        return {}



from typing import List, Optional, Tuple

from darwin.future.core.client import Client
from darwin.future.data_objects.team import Team, TeamMember


def get_team(client: Client, team_slug: Optional[str] = None) -> Team:
    """Returns the team with the given slug"""
    if not team_slug:
        team_slug = client.config.default_team
    response = client.get(f"/teams/{team_slug}/")
    return Team.parse_obj(response)


def get_team_members(client: Client) -> Tuple[List[TeamMember], List[Exception]]:
    response = client.get(f"/memberships")
    members = []
    errors = []
    for item in response:
        try:
            members.append(TeamMember.parse_obj(item))
        except Exception as e:
            errors.append(e)
    return (members, errors)



from pathlib import Path
from typing import Iterator, List

from darwin.datatypes import AnnotationFile, ExportParser, PathLike
from darwin.utils import parse_darwin_json, split_video_annotation


def darwin_to_dt_gen(file_paths: List[PathLike], split_sequences: bool) -> Iterator[AnnotationFile]:
    """
    Parses the given paths recursively and into an ``Iterator`` of ``AnnotationFile``\\s.

    Parameters
    ----------
    file_paths : List[PathLike]
        The paths of the files or directories we want to parse.

    split_sequences: bool
        When `True`, all videos will be split into individual frame images.

    Returns
    -------
    Iterator[AnnotationFile]
        An ``Iterator`` of the parsed ``AnnotationFile``\\s.
    """
    count = 0
    for file_path in map(Path, file_paths):
        files = file_path.glob("**/*") if file_path.is_dir() else [file_path]
        for f in files:
            if f.suffix != ".json":
                continue
            data = parse_darwin_json(f, count)
            if data:
                if data.is_video and split_sequences:
                    for d in split_video_annotation(data):
                        d.seq = count
                        count += 1
                        yield d
                else:
                    yield data
            count += 1


def export_annotations(
    exporter: ExportParser,
    file_paths: List[PathLike],
    output_directory: PathLike,
    split_sequences: bool = True,
) -> None:
    """
    Converts a set of files to a different annotation format.

    Parameters
    ----------
    exporter : ExportParser
        The parser to use.
    file_paths : List[PathLike]
        The files we want to parse.
    output_directory : PathLike
        Where the parsed files will be placed after the operation is complete.
    """
    print(f"Converting annotations...")
    exporter(
        darwin_to_dt_gen(file_paths, split_sequences=split_sequences),
        Path(output_directory),
    )
    print(f"Converted annotations saved at {output_directory}")



import colorsys
import os
from csv import writer as csv_writer
from functools import reduce
from pathlib import Path
from typing import Dict, Iterable, List, Literal, Optional, Set, Tuple, get_args

import numpy as np

try:
    from numpy.typing import NDArray
except ImportError:
    NDArray = Any  # type:ignore
from PIL import Image
from upolygon import draw_polygon

import darwin.datatypes as dt
from darwin.exceptions import DarwinException
from darwin.utils import convert_polygons_to_sequences, ispolygon


def get_palette(mode: dt.MaskTypes.Mode, categories: List[str]) -> dt.MaskTypes.Palette:
    """
    Returns a palette for the given mode and categories.

    Parameters
    ----------
    mode: dt.MaskTypes.Mode
        The mode to use for the palette.
    categories: List[str]
        A list of categories to be rendered.

    Returns
    -------
    dt.MaskTypes.Palette
        A dict of categories and their corresponding palette value.
    """

    if not mode in get_args(dt.MaskTypes.Mode):
        raise ValueError(f"Unknown mode {mode}.") from DarwinException

    if not isinstance(categories, list) or not categories:
        raise ValueError(f"categories must be a non-empty list. Got {categories}.")

    num_categories: int = len(categories)
    if mode == "index":
        if num_categories > 254:
            raise ValueError("maximum number of classes supported: 254.")
        palette = {c: i for i, c in enumerate(categories)}

    if mode == "grey":
        if num_categories > 254:
            raise ValueError("maximum number of classes supported: 254.")
        elif num_categories == 1:
            raise ValueError("only having the '__background__' class is not allowed. Please add more classes.")

        palette = {c: int(i * 255 / (num_categories - 1)) for i, c in enumerate(categories)}

    if mode == "rgb":
        if num_categories > 360:
            raise ValueError("maximum number of classes supported: 360.")
        palette = {c: i for i, c in enumerate(categories)}

    if not palette:
        raise ValueError(f"Failed to generate a palette.", mode, categories) from DarwinException

    return palette


def get_rgb_colours(categories: dt.MaskTypes.CategoryList) -> Tuple[dt.MaskTypes.RgbColors, dt.MaskTypes.RgbPalette]:
    """
    Returns a list of RGB colours and a dict of categories and their corresponding RGB palette value.

    Parameters
    ----------
    categories: dt.MaskTypes.CategoryList
        A list of categories to be rendered.

    Returns
    -------
    dt.MaskTypes.RgbColors
        A list of RGB colours for each category.
    dt.MaskTypes.RgbPalette
        A dict of categories and their corresponding RGB palette value.
    """
    num_categories: int = len(categories)

    # Generate HSV colours for all classes except for BG
    SATURATION_OF_COLOUR: float = 0.8
    VALUE_OF_COLOUR: float = 1.0
    hsv_colours: dt.MaskTypes.HsvColors = [
        (x / num_categories, SATURATION_OF_COLOUR, VALUE_OF_COLOUR) for x in range(num_categories - 1)
    ]
    rgb_colour_list: dt.MaskTypes.RgbColorList = list(
        map(lambda x: [int(e * 255) for e in colorsys.hsv_to_rgb(*x)], hsv_colours)
    )
    # Now we add BG class with [0 0 0] RGB value
    rgb_colour_list.insert(0, [0, 0, 0])
    palette_rgb: dt.MaskTypes.RgbPalette = {c: rgb for c, rgb in zip(categories, rgb_colour_list)}
    rgb_colours: dt.MaskTypes.RgbColors = [c for e in rgb_colour_list for c in e]

    return rgb_colours, palette_rgb


def get_render_mode(annotations: List[dt.AnnotationLike]) -> dt.MaskTypes.TypeOfRender:
    """
    Returns the type of render mode for the given annotations.

    Parameters
    ----------
    annotations: List[dt.AnnotationLike]
        A list of annotations to be rendered.

    Returns
    -------
    TypeOfRenderType
        A string reading either "raster" or "polygon".
    """
    non_video_annotations: List[dt.Annotation] = [a for a in annotations if not isinstance(a, dt.VideoAnnotation)]

    if not non_video_annotations:
        return "polygon"

    list_of_keys: List[str] = reduce(list.__add__, [list(a.data.keys()) for a in non_video_annotations])
    keys: Set[str] = set(list_of_keys)

    is_raster_mask = ("mask" in keys) and ("raster_layer" in keys)
    is_polygon = ("path" in keys) or ("paths" in keys)

    raster_layer_count = len([a for a in keys if a == "raster_layer"])

    if is_raster_mask and is_polygon:
        raise ValueError("Cannot have both raster and polygon annotations in the same file")

    if is_raster_mask and raster_layer_count > 1:
        raise ValueError("Cannot have more than one raster layer in the same file")

    if is_raster_mask:
        return "raster"

    if is_polygon:
        return "polygon"

    raise ValueError("No renderable annotations found in file, found keys: " + ",".join(keys))


def colours_in_rle(
    colours: dt.MaskTypes.ColoursDict,
    raster_layer: dt.RasterLayer,
    mask_lookup: Dict[str, dt.AnnotationMask],
) -> dt.MaskTypes.ColoursDict:
    """
    Returns a dictionary of colours for each mask in the given RLE.

    Parameters
    ----------
    colours: dt.MaskTypes.ColoursDict
        A dictionary of colours for each mask in the given RLE.
    mask_annotations: List[dt.AnnotationMask]
        A list of masks to get the colours for.
    mask_lookup: Set[str, dt.AnnotationMask]
        A lookup table for the masks.

    Returns
    -------
    dt.MaskTypes.ColoursDict
        A dictionary of colours for each mask in the given RLE.
    """
    for uuid, colour_value in raster_layer.mask_annotation_ids_mapping.items():
        mask: Optional[dt.AnnotationMask] = mask_lookup.get(uuid)

        if mask is None:
            raise ValueError(f"Could not find mask with uuid {uuid} in mask lookup table.")

        if not mask.name in colours:
            colours[mask.name] = colour_value

    return colours  # Returns same item as the outset, technically not needed, but best practice.


def rle_decode(rle: dt.MaskTypes.UndecodedRLE) -> List[int]:
    """Decodes a run-length encoded list of integers.

    Args:
        rle (List[int]): A run-length encoded list of integers.

    Returns:
        List[int]: The decoded list of integers.
    """
    if len(rle) % 2 != 0:
        raise ValueError("RLE must be a list of pairs of integers.")

    output = []
    for i in range(0, len(rle), 2):
        output += [rle[i]] * rle[i + 1]

    return output


def get_or_generate_colour(cat_name: str, colours: dt.MaskTypes.ColoursDict) -> int:
    """
    Returns the colour for the given category name, or generates a new one if it doesn't exist.

    Parameters
    ----------
    cat_name: str
        The name of the category.
    colours: dt.MaskTypes.ColoursDict
        A dictionary of category names and their corresponding colours.

    Returns
    -------
    int - the integer for the colour name.  These will later be reassigned to a wider spread across the colour spectrum.
    """
    if not cat_name in colours:
        colours[cat_name] = len(colours) + 1

    return colours[cat_name]


def render_polygons(
    mask: NDArray,
    colours: dt.MaskTypes.ColoursDict,
    categories: dt.MaskTypes.CategoryList,
    annotations: List[dt.AnnotationLike],
    annotation_file: dt.AnnotationFile,
    height: int,
    width: int,
) -> dt.MaskTypes.RendererReturn:
    """
    Renders the polygons in the given annotations onto the given mask.

    Parameters
    ----------
    mask: NDArray
        The mask to render the polygons onto.
    colours: dt.MaskTypes.ColoursDict
        A dictionary of category names and their corresponding colours.
    categories: dt.MaskTypes.CategoryList
        A list of category names.
    annotations: List[dt.AnnotationLike]
        A list of annotations to be rendered.
    annotation_file: dt.AnnotationFile
        The annotation file that the annotations belong to.
    height: int
        The height of the image.
    width: int
        The width of the image.

    Returns
    -------
    Tuple[List[Exception], Image, dt.MaskTypes.CategoryList, dt.MaskTypes.ColoursDict]
    """

    errors: List[Exception] = []

    for a in annotations:
        try:
            if isinstance(a, dt.VideoAnnotation):
                print(f"Skipping video annotation from file {annotation_file.filename}")
                continue

            cat = a.annotation_class.name
            if not cat in categories:
                categories.append(cat)

            if a.annotation_class.annotation_type == "polygon":
                polygon = a.data["path"]
            elif a.annotation_class.annotation_type == "complex_polygon":
                polygon = a.data["paths"]
            else:
                raise ValueError(f"Unknown annotation type {a.annotation_class.annotation_type}")
            sequence = convert_polygons_to_sequences(polygon, height=height, width=width)

            colour_to_draw = get_or_generate_colour(cat, colours)
            mask = draw_polygon(mask, sequence, colour_to_draw)

            if cat not in colours:
                colours[cat] = colour_to_draw

        except Exception as e:
            errors.append(e)
            continue

    # It's not necessary to return the mask, it's modified in place, but it's more explicit
    return errors, mask, categories, colours


def render_raster(
    mask: NDArray,
    colours: dt.MaskTypes.ColoursDict,
    categories: dt.MaskTypes.CategoryList,
    annotations: List[dt.AnnotationLike],
    annotation_file: dt.AnnotationFile,  # Not used, but kept for consistency
    height: int,
    width: int,
) -> dt.MaskTypes.RendererReturn:
    """
    Renders the raster layers in the given annotations onto the given mask.

    Parameters
    ----------
    mask: NDArray
        The mask to render the polygons onto. Not used.  Only returned if no errors occur.
    colours: dt.MaskTypes.ColoursDict
        The colours list. Only returned if no errors occur.
    annotations: List[dt.AnnotationLike]
        A list of annotations to be rendered.
    annotation_file: dt.AnnotationFile
        Not used. Present for interface consistency.
    height: int
        The height of the image.
    width: int
        The width of the image.

    Returns
    -------
    Tuple[List[Exception], Image, dt.MaskTypes.CategoryList, dt.MaskTypes.ColoursDict]
    """
    errors: List[Exception] = []

    mask_annotations: List[dt.AnnotationMask] = []
    raster_layer: Optional[dt.RasterLayer] = None

    mask_lookup: Dict[str, dt.AnnotationMask] = dict()

    for a in annotations:
        if isinstance(a, dt.VideoAnnotation):
            continue

        data = a.data

        if "mask" in data:
            new_mask = dt.AnnotationMask(
                id=data["id"],
                name=data["name"],
                slot_names=a.slot_names,
            )
            try:
                new_mask.validate()
            except Exception as e:
                errors.append(e)
                continue

            mask_annotations.append(new_mask)

            if not new_mask.id in mask_lookup:
                mask_lookup[new_mask.id] = new_mask

            # Add the category to the list of categories
            if not new_mask.name in categories:
                categories.append(new_mask.name)

        if "raster_layer" in data and (rl := data["raster_layer"]):

            if raster_layer:
                errors.append(ValueError(f"Annotation {a.id} has more than one raster layer"))
                break

            new_rl = dt.RasterLayer(
                rle=rl["dense_rle"],
                decoded=rle_decode(rl["dense_rle"]),  # type: ignore
                slot_names=a.slot_names,
                mask_annotation_ids_mapping=rl["mask_annotation_ids_mapping"],
                total_pixels=rl["total_pixels"],
            )
            new_rl.validate()
            raster_layer = new_rl

    if not raster_layer:
        errors.append(ValueError(f"Annotation has no raster layer"))
        return errors, mask, categories, colours

    if not mask_annotations:
        errors.append(ValueError(f"Annotation has no masks"))
        return errors, mask, categories, colours

    try:
        colours = colours_in_rle(colours, raster_layer, mask_lookup)
    except Exception as e:
        errors.append(e)

    mask = np.array(raster_layer.decoded, dtype=np.uint8).reshape(height, width)

    return errors, mask, categories, colours


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path, mode: dt.MaskTypes.Mode) -> None:
    masks_dir: Path = output_dir / "masks"
    masks_dir.mkdir(exist_ok=True, parents=True)
    annotation_files = list(annotation_files)

    categories: List[str] = ["__background__"]
    colours: dt.MaskTypes.ColoursDict = dict()

    for annotation_file in annotation_files:

        image_rel_path = os.path.splitext(annotation_file.full_path)[0].lstrip("/")
        outfile = masks_dir / f"{image_rel_path}.png"
        outfile.parent.mkdir(parents=True, exist_ok=True)

        height = annotation_file.image_height
        width = annotation_file.image_width
        if height is None or width is None:
            raise ValueError(f"Annotation file {annotation_file.filename} references an image with no height or width")

        mask: NDArray = np.zeros((height, width)).astype(np.uint8)
        annotations: List[dt.AnnotationLike] = [
            a
            for a in annotation_file.annotations
            if a.annotation_class.annotation_type in ["polygon", "complex_polygon", "raster_layer", "mask"]
        ]

        render_type = get_render_mode(annotations)

        if render_type == "raster":
            # Add categories to list
            errors, mask, categories, colours = render_raster(
                mask, colours, categories, annotations, annotation_file, height, width
            )

        else:
            #  Add categories to list
            errors, mask, categories, colours = render_polygons(
                mask, colours, categories, annotations, annotation_file, height, width
            )

        if errors:
            print(f"Errors rendering {annotation_file.filename}:")
            for e in errors:
                print(e)

            raise DarwinException.from_multiple_exceptions(errors)

        # Map to palette
        mask = np.array(mask, dtype=np.uint8)  # Final double check that type is using correct dtype

        palette = get_palette(mode, categories)
        if mode == "rgb":
            rgb_colours, palette_rgb = get_rgb_colours(categories)
            image = Image.fromarray(mask, "P")
            image.putpalette(rgb_colours)
            image = image.convert("RGB")
        elif mode == "grey":
            for value, colour in enumerate(palette.values()):
                mask = np.where(mask == value, colour, mask)
            image = Image.fromarray(mask)
        else:
            image = Image.fromarray(mask)
        image.save(outfile)

    with open(output_dir / "class_mapping.csv", "w") as f:
        writer = csv_writer(f)
        writer.writerow(["class_name", "class_color"])

        for c in categories:
            if mode == "rgb":
                writer.writerow([c, f"{palette_rgb[c][0]} {palette_rgb[c][1]} {palette_rgb[c][2]}"])
            else:
                writer.writerow([c, f"{palette[c]}"])



from typing import Any, List, Optional

from darwin.future.data_objects.typing import KeyValuePairDict, UnknownType


class DarwinException(Exception):
    """
    Generic Darwin exception.

    Used to differentiate from errors that originate in our code, and those that originate in
    third-party libraries.

    Extends `Exception` and adds a `parent_exception` field to store the original exception.

    Also has a `combined_exceptions` field to store a list of exceptions that were combined into
    """

    parent_exception: Optional[Exception] = None
    combined_exceptions: Optional[List[Exception]] = None

    def __init__(self, *args: UnknownType, **kwargs: KeyValuePairDict) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def from_exception(cls, exc: Exception) -> "DarwinException":
        """
        Creates a new exception from an existing exception.

        Parameters
        ----------
        exc: Exception
            The existing exception.

        Returns
        -------
        DarwinException
            The new exception.
        """
        instance = cls(str(exc))
        instance.parent_exception = exc

        return instance

    def __str__(self) -> str:
        output_string = f"{self.__class__.__name__}: {super().__str__()}\n"
        if self.parent_exception:
            output_string += f"Parent exception: {self.parent_exception}\n"

        if self.combined_exceptions:
            output_string += f"Combined exceptions: {self.combined_exceptions}\n"

        return output_string

    def __repr__(self) -> str:
        return super().__repr__()


class ValidationError(DarwinException):
    pass



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, overload
from urllib.parse import urlparse

import requests
import yaml
from pydantic import BaseModel, root_validator, validator
from requests.adapters import HTTPAdapter, Retry

from darwin.future.core.types.query import Query
from darwin.future.exceptions.client import NotFound, Unauthorized

JSONType = Dict[str, Any]  # type: ignore


class TeamsConfig(BaseModel):
    api_key: str
    datasets_dir: Path


class DarwinConfig(BaseModel):
    """Configuration object for the client

    Attributes
    ----------
    api_key: Optional[str], api key to authenticate
    base_url: pydantic.HttpUrl, base url of the API
    default_team: Optional[Team], default team to make requests to
    """

    api_key: Optional[str]
    datasets_dir: Optional[Path]
    api_endpoint: str
    base_url: str
    default_team: str
    teams: Dict[str, TeamsConfig]

    @validator("base_url")
    def validate_base_url(cls, v: str) -> str:
        v = v.strip()
        if not v.endswith("/"):
            v += "/"
        check = urlparse(v)
        assert check.scheme in {"http", "https"}, "base_url must start with http or https"
        assert check.netloc, "base_url must contain a domain"
        return v

    @root_validator(pre=True)
    def remove_global(cls, values: dict) -> dict:
        if "global" not in values:
            return values
        global_conf = values["global"]
        del values["global"]
        values.update(global_conf)
        return values

    @root_validator()
    def validate_defaults(cls, values: dict) -> dict:
        if values["api_key"]:
            return values
        assert values["default_team"] in values["teams"]
        team = values["default_team"]
        values["api_key"] = values["teams"][team].api_key
        values["datasets_dir"] = values["teams"][team].datasets_dir
        return values

    @staticmethod
    def local() -> DarwinConfig:
        return DarwinConfig.from_file(Path.home() / ".darwin" / "config.yaml")

    @staticmethod
    def from_file(path: Path) -> DarwinConfig:
        if path.suffix.lower() == ".yaml":
            data = DarwinConfig._parse_yaml(path)
            return DarwinConfig.parse_obj(data)
        else:
            return DarwinConfig.parse_file(path)

    @staticmethod
    def _parse_yaml(path: Path) -> dict:
        with open(path, encoding="utf-8") as infile:
            data = yaml.safe_load(infile)
        return data

    class Config:
        validate_assignment = True


class Result(BaseModel):
    """Default model for a result, to be extended by other models specific to the API"""

    ...


class PageDetail(BaseModel):
    """Page details model for managing pagination

    Attributes
    ----------
    count: int, current position
    next: Optional[str], url for the next page
    previous: Optional[str], url for the previous page
    """

    count: int
    next: Optional[str]
    previous: Optional[str]


class Page(BaseModel):
    """Page of results

    Attributes
    ----------
    results: List[Result], list of results
    detail: PageDetail, details about the page
    """

    results: List[Result]
    detail: PageDetail


class Cursor(ABC):
    """Abstract class for a cursor

    Attributes
    ----------
    url: str, url of the endpoint
    client: Client, client used to make requests
    """

    def __init__(self, url: str, client: Client):
        self.url = url
        self.client = client
        self.current_page: Optional[Page] = None

    @abstractmethod
    def execute(self, query: Query) -> Page:
        pass

    @abstractmethod
    def __iter__(self) -> Page:
        pass

    @abstractmethod
    def __next__(self) -> Page:
        pass


class Client:
    """Client Object to manage and make requests to the Darwin API
    Attributes
    ----------
    url: str, url of the endpoint
    api_key: str, api key to authenticate
    team: Team, team to make requests to
    """

    def __init__(self, config: DarwinConfig, retries: Optional[Retry] = None) -> None:
        self.config = config
        self.session = requests.Session()
        if not retries:
            retries = Retry(total=3, backoff_factor=0.2, status_forcelist=[500, 502, 503, 504])
        self._setup_session(retries)
        self._mappings = {
            "get": self.session.get,
            "put": self.session.put,
            "post": self.session.post,
            "delete": self.session.delete,
            "patch": self.session.patch,
        }

    @classmethod
    def local(cls) -> Client:
        config = DarwinConfig.local()
        return Client(config)

    def _setup_session(self, retries: Retry) -> None:
        self.session.headers.update(self.headers)
        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

    @property
    def headers(self) -> Dict[str, str]:
        http_headers: Dict[str, str] = {"Content-Type": "application/json"}
        if self.config.api_key:
            http_headers["Authorization"] = f"ApiKey {self.config.api_key}"
        return http_headers

    @overload
    def _generic_call(self, method: Callable[[str], requests.Response], endpoint: str) -> dict:
        ...

    @overload
    def _generic_call(self, method: Callable[[str, dict], requests.Response], endpoint: str, payload: dict) -> dict:
        ...

    def _generic_call(self, method: Callable, endpoint: str, payload: Optional[dict] = None) -> JSONType:
        endpoint = self._sanitize_endpoint(endpoint)
        url = self.config.api_endpoint + endpoint
        if payload is not None:
            response = method(url, payload)
        else:
            response = method(url)

        raise_for_darwin_exception(response)
        response.raise_for_status()

        return response.json()

    def cursor(self) -> Cursor:
        pass

    def get(self, endpoint: str) -> JSONType:
        return self._generic_call(self.session.get, endpoint)

    def put(self, endpoint: str, data: dict) -> JSONType:
        return self._generic_call(self.session.put, endpoint, data)

    def post(self, endpoint: str, data: dict) -> JSONType:
        return self._generic_call(self.session.post, endpoint, data)

    def delete(self, endpoint: str) -> JSONType:
        return self._generic_call(self.session.delete, endpoint)

    def patch(self, endpoint: str, data: dict) -> JSONType:
        return self._generic_call(self.session.patch, endpoint, data)

    def _sanitize_endpoint(self, endpoint: str) -> str:
        return endpoint.strip().strip("/")


def raise_for_darwin_exception(response: requests.Response) -> None:
    """Raises an exception if the response.status_code matches a known darwin error

    Parameters
    ----------
    response: requests.Response, response to check
    """
    if response.status_code == 200:
        return
    if response.status_code == 401:
        raise Unauthorized(response)
    if response.status_code == 404:
        raise NotFound(response)
